"""User template storage.

Handles saving and loading user-created templates to the local filesystem.
"""

from __future__ import annotations

import logging
from datetime import datetime
from pathlib import Path

import yaml

from gaol.config import get_config_dir
from gaol.templates.exceptions import (
    TemplateExistsError,
    TemplateNotFoundError,
    TemplateParseError,
    TemplateStoreError,
)
from gaol.templates.models import Template, TemplateMetadata

logger = logging.getLogger(__name__)


class TemplateStore:
    """Persistent storage for user-created templates.

    Templates are stored as YAML files in ~/.config/gaol/templates/
    with the structure {category}/{name}.yaml matching the template ID.
    """

    def __init__(self, store_dir: Path | None = None) -> None:
        """Initialize the template store.

        Args:
            store_dir: Directory to store templates. Defaults to
                       ~/.config/gaol/templates/
        """
        if store_dir is None:
            store_dir = get_config_dir() / "templates"
        self._store_dir = store_dir
        self._store_dir.mkdir(parents=True, exist_ok=True)

    def _get_path(self, template_id: str) -> Path:
        """Get the path for a template file.

        Args:
            template_id: Template ID (e.g., 'custom/my-template')

        Returns:
            Path to the template YAML file
        """
        # Template ID is category/name
        parts = template_id.split("/", 1)
        if len(parts) == 2:
            category, name = parts
            category_dir = self._store_dir / category
            category_dir.mkdir(parents=True, exist_ok=True)
            return category_dir / f"{name}.yaml"
        else:
            return self._store_dir / f"{template_id}.yaml"

    def save(
        self,
        template: Template,
        overwrite: bool = False,
    ) -> Path:
        """Save a template to the store.

        Args:
            template: Template to save
            overwrite: Whether to overwrite existing template

        Returns:
            Path to the saved template file

        Raises:
            TemplateExistsError: If template exists and overwrite is False
            TemplateStoreError: If save fails
        """
        path = self._get_path(template.metadata.id)

        if path.exists() and not overwrite:
            raise TemplateExistsError(template.metadata.id, str(path))

        # Update timestamp
        template.metadata.updated_at = datetime.now()
        if template.metadata.created_at is None:
            template.metadata.created_at = template.metadata.updated_at

        try:
            data = template.to_dict()
            content = yaml.dump(data, default_flow_style=False, sort_keys=False)
            path.write_text(content)
            logger.debug(f"Saved template to {path}")
            return path
        except Exception as e:
            raise TemplateStoreError("save", str(e)) from e

    def get(self, template_id: str) -> Template | None:
        """Get a template from the store.

        Args:
            template_id: Template ID

        Returns:
            Template if found, None otherwise
        """
        path = self._get_path(template_id)
        if not path.exists():
            return None

        try:
            content = path.read_text()
            data = yaml.safe_load(content)
            return Template.from_dict(data)
        except Exception as e:
            logger.warning(f"Failed to load template {template_id}: {e}")
            return None

    def delete(self, template_id: str) -> bool:
        """Delete a template from the store.

        Args:
            template_id: Template ID

        Returns:
            True if deleted, False if not found
        """
        path = self._get_path(template_id)
        if not path.exists():
            return False

        try:
            path.unlink()
            logger.debug(f"Deleted template {template_id}")

            # Clean up empty category directory
            category_dir = path.parent
            if category_dir != self._store_dir and not any(category_dir.iterdir()):
                category_dir.rmdir()

            return True
        except Exception as e:
            logger.error(f"Failed to delete template {template_id}: {e}")
            return False

    def list(self, category: str | None = None) -> list[Template]:
        """List all templates in the store.

        Args:
            category: Filter by category (optional)

        Returns:
            List of templates
        """
        templates = []

        if category:
            # List templates in specific category
            category_dir = self._store_dir / category
            if category_dir.exists():
                for path in category_dir.glob("*.yaml"):
                    template_id = f"{category}/{path.stem}"
                    template = self.get(template_id)
                    if template:
                        templates.append(template)
        else:
            # List all templates
            for category_dir in self._store_dir.iterdir():
                if category_dir.is_dir():
                    for path in category_dir.glob("*.yaml"):
                        template_id = f"{category_dir.name}/{path.stem}"
                        template = self.get(template_id)
                        if template:
                            templates.append(template)

            # Also check for templates in root (no category)
            for path in self._store_dir.glob("*.yaml"):
                template_id = path.stem
                template = self.get(template_id)
                if template:
                    templates.append(template)

        return templates

    def list_ids(self, category: str | None = None) -> list[str]:
        """List all template IDs in the store.

        Args:
            category: Filter by category (optional)

        Returns:
            List of template IDs
        """
        ids = []

        if category:
            category_dir = self._store_dir / category
            if category_dir.exists():
                for path in category_dir.glob("*.yaml"):
                    ids.append(f"{category}/{path.stem}")
        else:
            for category_dir in self._store_dir.iterdir():
                if category_dir.is_dir():
                    for path in category_dir.glob("*.yaml"):
                        ids.append(f"{category_dir.name}/{path.stem}")
            # Root templates
            for path in self._store_dir.glob("*.yaml"):
                ids.append(path.stem)

        return sorted(ids)

    def exists(self, template_id: str) -> bool:
        """Check if a template exists in the store.

        Args:
            template_id: Template ID

        Returns:
            True if template exists
        """
        return self._get_path(template_id).exists()

    def get_categories(self) -> list[str]:
        """Get all categories in the store.

        Returns:
            List of category names
        """
        categories = []
        for path in self._store_dir.iterdir():
            if path.is_dir() and any(path.glob("*.yaml")):
                categories.append(path.name)
        return sorted(categories)


# Module-level store instance
_template_store: TemplateStore | None = None


def get_template_store() -> TemplateStore:
    """Get the global template store instance.

    Returns:
        TemplateStore instance
    """
    global _template_store
    if _template_store is None:
        _template_store = TemplateStore()
    return _template_store



"""Firewall rule management for port forwarding."""

from __future__ import annotations

import shutil
import subprocess
from enum import Enum
from typing import NamedTuple

from gaol.networking.models import Protocol


class FirewallBackend(str, Enum):
    """Firewall backend type."""

    IPTABLES = "iptables"
    NFTABLES = "nftables"
    AUTO = "auto"


class FirewallError(Exception):
    """Firewall operation error."""

    pass


class ForwardingRule(NamedTuple):
    """A port forwarding rule."""

    host_port: int
    guest_ip: str
    guest_port: int
    protocol: Protocol
    host_ip: str = "0.0.0.0"
    comment: str = ""


# Chain/table names for gaol rules
GAOL_CHAIN = "GAOL"
GAOL_NFT_TABLE = "gaol"


def detect_firewall_backend() -> FirewallBackend:
    """Detect the available firewall backend.

    Returns:
        FirewallBackend that is available

    Raises:
        FirewallError: If no firewall backend is available
    """
    # Prefer nftables if available and active
    if shutil.which("nft"):
        result = subprocess.run(
            ["nft", "list", "tables"],
            capture_output=True,
        )
        if result.returncode == 0:
            return FirewallBackend.NFTABLES

    if shutil.which("iptables"):
        return FirewallBackend.IPTABLES

    raise FirewallError("No firewall backend available (need iptables or nftables)")


def _run_cmd(
    cmd: list[str], use_sudo: bool = True
) -> subprocess.CompletedProcess[str]:
    """Run a command with optional sudo."""
    if use_sudo:
        cmd = ["sudo"] + cmd
    return subprocess.run(cmd, capture_output=True, text=True)


class FirewallManager:
    """Manages firewall rules for port forwarding."""

    def __init__(
        self,
        backend: FirewallBackend = FirewallBackend.AUTO,
        use_sudo: bool = True,
    ) -> None:
        self.use_sudo = use_sudo
        self._setup_done = False

        if backend == FirewallBackend.AUTO:
            self.backend = detect_firewall_backend()
        else:
            self.backend = backend

    def setup(self) -> None:
        """Set up firewall chains/tables for gaol.

        Creates the necessary chains/tables if they don't exist.
        """
        if self._setup_done:
            return

        if self.backend == FirewallBackend.NFTABLES:
            self._setup_nftables()
        else:
            self._setup_iptables()

        self._setup_done = True

    def _setup_nftables(self) -> None:
        """Set up nftables table and chains."""
        # Create table if not exists
        _run_cmd(["nft", "add", "table", "ip", GAOL_NFT_TABLE], self.use_sudo)

        # Create prerouting chain for DNAT
        _run_cmd(
            [
                "nft",
                "add",
                "chain",
                "ip",
                GAOL_NFT_TABLE,
                "prerouting",
                "{ type nat hook prerouting priority -100 ; }",
            ],
            self.use_sudo,
        )

        # Create postrouting chain for SNAT/masquerade
        _run_cmd(
            [
                "nft",
                "add",
                "chain",
                "ip",
                GAOL_NFT_TABLE,
                "postrouting",
                "{ type nat hook postrouting priority 100 ; }",
            ],
            self.use_sudo,
        )

        # Create filter chain for forwarding
        _run_cmd(
            [
                "nft",
                "add",
                "chain",
                "ip",
                GAOL_NFT_TABLE,
                "forward",
                "{ type filter hook forward priority 0 ; }",
            ],
            self.use_sudo,
        )

    def _setup_iptables(self) -> None:
        """Set up iptables chains."""
        # Create custom chain in nat table
        result = _run_cmd(
            ["iptables", "-t", "nat", "-N", GAOL_CHAIN], self.use_sudo
        )

        if result.returncode == 0:
            # Chain created, add jump rules
            _run_cmd(
                ["iptables", "-t", "nat", "-A", "PREROUTING", "-j", GAOL_CHAIN],
                self.use_sudo,
            )
            _run_cmd(
                ["iptables", "-t", "nat", "-A", "OUTPUT", "-j", GAOL_CHAIN],
                self.use_sudo,
            )

        # Create custom chain in filter table for forwarding
        _run_cmd(["iptables", "-N", f"{GAOL_CHAIN}_FWD"], self.use_sudo)
        _run_cmd(
            ["iptables", "-A", "FORWARD", "-j", f"{GAOL_CHAIN}_FWD"],
            self.use_sudo,
        )

    def add_forwarding_rule(self, rule: ForwardingRule) -> None:
        """Add a port forwarding rule.

        Args:
            rule: Forwarding rule to add

        Raises:
            FirewallError: If rule creation fails
        """
        self.setup()

        if self.backend == FirewallBackend.NFTABLES:
            self._add_nft_rule(rule)
        else:
            self._add_iptables_rule(rule)

    def _add_nft_rule(self, rule: ForwardingRule) -> None:
        """Add forwarding rule using nftables."""
        proto = rule.protocol.value
        if rule.protocol == Protocol.BOTH:
            # Add both TCP and UDP rules
            for p in [Protocol.TCP, Protocol.UDP]:
                self._add_nft_rule(
                    ForwardingRule(
                        rule.host_port,
                        rule.guest_ip,
                        rule.guest_port,
                        p,
                        rule.host_ip,
                        rule.comment,
                    )
                )
            return

        # DNAT rule (prerouting)
        dnat_rule = [
            "nft",
            "add",
            "rule",
            "ip",
            GAOL_NFT_TABLE,
            "prerouting",
            proto,
            "dport",
            str(rule.host_port),
            "dnat",
            "to",
            f"{rule.guest_ip}:{rule.guest_port}",
        ]
        if rule.comment:
            dnat_rule.extend(["comment", f'"{rule.comment}"'])

        result = _run_cmd(dnat_rule, self.use_sudo)
        if result.returncode != 0:
            raise FirewallError(f"Failed to add DNAT rule: {result.stderr}")

        # Forward allow rule
        fwd_rule = [
            "nft",
            "add",
            "rule",
            "ip",
            GAOL_NFT_TABLE,
            "forward",
            "ip",
            "daddr",
            rule.guest_ip,
            proto,
            "dport",
            str(rule.guest_port),
            "accept",
        ]
        result = _run_cmd(fwd_rule, self.use_sudo)
        if result.returncode != 0:
            raise FirewallError(f"Failed to add forward rule: {result.stderr}")

    def _add_iptables_rule(self, rule: ForwardingRule) -> None:
        """Add forwarding rule using iptables."""
        proto = rule.protocol.value
        if rule.protocol == Protocol.BOTH:
            for p in [Protocol.TCP, Protocol.UDP]:
                self._add_iptables_rule(
                    ForwardingRule(
                        rule.host_port,
                        rule.guest_ip,
                        rule.guest_port,
                        p,
                        rule.host_ip,
                        rule.comment,
                    )
                )
            return

        # DNAT rule
        dnat_cmd = [
            "iptables",
            "-t",
            "nat",
            "-A",
            GAOL_CHAIN,
            "-p",
            proto,
            "--dport",
            str(rule.host_port),
            "-j",
            "DNAT",
            "--to-destination",
            f"{rule.guest_ip}:{rule.guest_port}",
        ]
        if rule.host_ip != "0.0.0.0":
            dnat_cmd.extend(["-d", rule.host_ip])
        if rule.comment:
            dnat_cmd.extend(["-m", "comment", "--comment", rule.comment])

        result = _run_cmd(dnat_cmd, self.use_sudo)
        if result.returncode != 0:
            raise FirewallError(f"Failed to add DNAT rule: {result.stderr}")

        # Forward allow rule
        fwd_cmd = [
            "iptables",
            "-A",
            f"{GAOL_CHAIN}_FWD",
            "-d",
            rule.guest_ip,
            "-p",
            proto,
            "--dport",
            str(rule.guest_port),
            "-j",
            "ACCEPT",
        ]
        result = _run_cmd(fwd_cmd, self.use_sudo)
        if result.returncode != 0:
            raise FirewallError(f"Failed to add forward rule: {result.stderr}")

    def remove_forwarding_rule(self, rule: ForwardingRule) -> None:
        """Remove a port forwarding rule.

        Args:
            rule: Forwarding rule to remove
        """
        if self.backend == FirewallBackend.NFTABLES:
            self._remove_nft_rule(rule)
        else:
            self._remove_iptables_rule(rule)

    def _remove_nft_rule(self, rule: ForwardingRule) -> None:
        """Remove nftables rule by pattern matching."""
        proto = rule.protocol.value
        if rule.protocol == Protocol.BOTH:
            for p in [Protocol.TCP, Protocol.UDP]:
                self._remove_nft_rule(
                    ForwardingRule(
                        rule.host_port,
                        rule.guest_ip,
                        rule.guest_port,
                        p,
                        rule.host_ip,
                        rule.comment,
                    )
                )
            return

        # List rules and find handle for prerouting
        result = _run_cmd(
            [
                "nft",
                "-a",
                "list",
                "chain",
                "ip",
                GAOL_NFT_TABLE,
                "prerouting",
            ],
            self.use_sudo,
        )

        if result.returncode == 0:
            for line in result.stdout.splitlines():
                if (
                    f"dport {rule.host_port}" in line
                    and proto in line
                    and "handle" in line
                ):
                    # Extract handle number
                    handle = line.split("handle")[-1].strip()
                    _run_cmd(
                        [
                            "nft",
                            "delete",
                            "rule",
                            "ip",
                            GAOL_NFT_TABLE,
                            "prerouting",
                            "handle",
                            handle,
                        ],
                        self.use_sudo,
                    )

        # Remove forward rule
        result = _run_cmd(
            ["nft", "-a", "list", "chain", "ip", GAOL_NFT_TABLE, "forward"],
            self.use_sudo,
        )

        if result.returncode == 0:
            for line in result.stdout.splitlines():
                if (
                    f"dport {rule.guest_port}" in line
                    and rule.guest_ip in line
                    and "handle" in line
                ):
                    handle = line.split("handle")[-1].strip()
                    _run_cmd(
                        [
                            "nft",
                            "delete",
                            "rule",
                            "ip",
                            GAOL_NFT_TABLE,
                            "forward",
                            "handle",
                            handle,
                        ],
                        self.use_sudo,
                    )

    def _remove_iptables_rule(self, rule: ForwardingRule) -> None:
        """Remove iptables rule."""
        proto = rule.protocol.value
        if rule.protocol == Protocol.BOTH:
            for p in [Protocol.TCP, Protocol.UDP]:
                self._remove_iptables_rule(
                    ForwardingRule(
                        rule.host_port,
                        rule.guest_ip,
                        rule.guest_port,
                        p,
                        rule.host_ip,
                        rule.comment,
                    )
                )
            return

        # Remove DNAT rule
        _run_cmd(
            [
                "iptables",
                "-t",
                "nat",
                "-D",
                GAOL_CHAIN,
                "-p",
                proto,
                "--dport",
                str(rule.host_port),
                "-j",
                "DNAT",
                "--to-destination",
                f"{rule.guest_ip}:{rule.guest_port}",
            ],
            self.use_sudo,
        )

        # Remove forward rule
        _run_cmd(
            [
                "iptables",
                "-D",
                f"{GAOL_CHAIN}_FWD",
                "-d",
                rule.guest_ip,
                "-p",
                proto,
                "--dport",
                str(rule.guest_port),
                "-j",
                "ACCEPT",
            ],
            self.use_sudo,
        )

    def clear_rules_for_vm(self, vm_ip: str) -> None:
        """Remove all forwarding rules for a VM.

        Args:
            vm_ip: IP address of the VM
        """
        if self.backend == FirewallBackend.NFTABLES:
            # List and remove all rules targeting this IP
            for chain in ["prerouting", "forward"]:
                result = _run_cmd(
                    ["nft", "-a", "list", "chain", "ip", GAOL_NFT_TABLE, chain],
                    self.use_sudo,
                )
                if result.returncode == 0:
                    for line in result.stdout.splitlines():
                        if vm_ip in line and "handle" in line:
                            handle = line.split("handle")[-1].strip()
                            _run_cmd(
                                [
                                    "nft",
                                    "delete",
                                    "rule",
                                    "ip",
                                    GAOL_NFT_TABLE,
                                    chain,
                                    "handle",
                                    handle,
                                ],
                                self.use_sudo,
                            )
        else:
            # For iptables, list rules with line numbers and delete
            # List DNAT rules
            result = _run_cmd(
                ["iptables", "-t", "nat", "-L", GAOL_CHAIN, "-n", "--line-numbers"],
                self.use_sudo,
            )
            if result.returncode == 0:
                # Parse in reverse to avoid line number shifting
                lines_to_delete = []
                for line in result.stdout.splitlines():
                    if vm_ip in line:
                        parts = line.split()
                        if parts and parts[0].isdigit():
                            lines_to_delete.append(int(parts[0]))

                for line_num in reversed(lines_to_delete):
                    _run_cmd(
                        [
                            "iptables",
                            "-t",
                            "nat",
                            "-D",
                            GAOL_CHAIN,
                            str(line_num),
                        ],
                        self.use_sudo,
                    )

            # List forward rules
            result = _run_cmd(
                ["iptables", "-L", f"{GAOL_CHAIN}_FWD", "-n", "--line-numbers"],
                self.use_sudo,
            )
            if result.returncode == 0:
                lines_to_delete = []
                for line in result.stdout.splitlines():
                    if vm_ip in line:
                        parts = line.split()
                        if parts and parts[0].isdigit():
                            lines_to_delete.append(int(parts[0]))

                for line_num in reversed(lines_to_delete):
                    _run_cmd(
                        [
                            "iptables",
                            "-D",
                            f"{GAOL_CHAIN}_FWD",
                            str(line_num),
                        ],
                        self.use_sudo,
                    )

    def list_rules(self) -> list[ForwardingRule]:
        """List all active forwarding rules.

        Returns:
            List of ForwardingRule objects
        """
        rules: list[ForwardingRule] = []

        if self.backend == FirewallBackend.NFTABLES:
            result = _run_cmd(
                ["nft", "list", "chain", "ip", GAOL_NFT_TABLE, "prerouting"],
                self.use_sudo,
            )
            # Parse nftables output (simplified parsing)
            if result.returncode == 0:
                for line in result.stdout.splitlines():
                    if "dnat to" in line:
                        # Try to parse: tcp dport 8080 dnat to 192.168.1.2:80
                        parts = line.strip().split()
                        try:
                            proto_idx = (
                                parts.index("tcp")
                                if "tcp" in parts
                                else parts.index("udp")
                            )
                            proto = Protocol(parts[proto_idx])
                            dport_idx = parts.index("dport") + 1
                            host_port = int(parts[dport_idx])
                            dnat_idx = parts.index("to") + 1
                            dest = parts[dnat_idx]
                            guest_ip, guest_port = dest.split(":")
                            rules.append(
                                ForwardingRule(
                                    host_port=host_port,
                                    guest_ip=guest_ip,
                                    guest_port=int(guest_port),
                                    protocol=proto,
                                )
                            )
                        except (ValueError, IndexError):
                            continue
        else:
            result = _run_cmd(
                ["iptables", "-t", "nat", "-L", GAOL_CHAIN, "-n"],
                self.use_sudo,
            )
            if result.returncode == 0:
                for line in result.stdout.splitlines():
                    if "DNAT" in line:
                        parts = line.split()
                        try:
                            proto = Protocol(parts[1].lower())
                            # Find dpt: and to:
                            for part in parts:
                                if part.startswith("dpt:"):
                                    host_port = int(part.split(":")[1])
                                if part.startswith("to:"):
                                    dest = part.split(":")[1:]
                                    guest_ip = dest[0]
                                    guest_port = int(dest[1]) if len(dest) > 1 else host_port
                            rules.append(
                                ForwardingRule(
                                    host_port=host_port,
                                    guest_ip=guest_ip,
                                    guest_port=guest_port,
                                    protocol=proto,
                                )
                            )
                        except (ValueError, IndexError):
                            continue

        return rules


# Convenience functions


def add_port_forward(
    host_port: int,
    guest_ip: str,
    guest_port: int,
    protocol: Protocol = Protocol.TCP,
    comment: str = "",
) -> None:
    """Add a port forwarding rule.

    Convenience function that creates a FirewallManager and adds the rule.
    """
    manager = FirewallManager()
    manager.add_forwarding_rule(
        ForwardingRule(
            host_port=host_port,
            guest_ip=guest_ip,
            guest_port=guest_port,
            protocol=protocol,
            comment=comment,
        )
    )


def remove_port_forward(
    host_port: int,
    guest_ip: str,
    guest_port: int,
    protocol: Protocol = Protocol.TCP,
) -> None:
    """Remove a port forwarding rule.

    Convenience function that creates a FirewallManager and removes the rule.
    """
    manager = FirewallManager()
    manager.remove_forwarding_rule(
        ForwardingRule(
            host_port=host_port,
            guest_ip=guest_ip,
            guest_port=guest_port,
            protocol=protocol,
        )
    )



"""Tests for profile loading and application."""

import pytest

from gaol.models.container import ContainerConfig, NetworkConfig
from gaol.models.profile import Profile
from gaol.profiles.applicator import MergedProfile, ProfileApplicator, get_profile_applicator
from gaol.profiles.loader import BUILTIN_PROFILES, ProfileLoader, get_profile_loader


class TestBuiltinProfiles:
    """Tests for built-in profiles."""

    def test_base_profile_exists(self) -> None:
        """Test that base profile is defined."""
        assert "base" in BUILTIN_PROFILES
        assert BUILTIN_PROFILES["base"]["name"] == "base"

    def test_all_builtin_profiles_valid(self) -> None:
        """Test that all built-in profiles can be loaded."""
        loader = ProfileLoader()
        for name in BUILTIN_PROFILES:
            profile = loader.get(name)
            assert profile.name == name

    def test_tailscale_profile(self) -> None:
        """Test tailscale profile configuration."""
        loader = ProfileLoader()
        profile = loader.get("tailscale")
        assert profile.requires_root is True
        assert "NET_ADMIN" in profile.capabilities_add
        assert "tailscaled" in profile.services

    def test_gui_profile(self) -> None:
        """Test gui profile configuration."""
        loader = ProfileLoader()
        profile = loader.get("gui")
        assert "DISPLAY" in profile.environment
        assert "/tmp/.X11-unix" in profile.volumes

    def test_dev_python_profile(self) -> None:
        """Test dev-python profile configuration."""
        loader = ProfileLoader()
        profile = loader.get("dev-python")
        assert "python3" in profile.packages
        assert "git" in profile.packages


class TestProfileLoader:
    """Tests for ProfileLoader class."""

    def test_list_available(self) -> None:
        """Test listing available profiles."""
        loader = ProfileLoader()
        available = loader.list_available()
        assert "base" in available
        assert "tailscale" in available
        assert "gui" in available

    def test_get_nonexistent_profile(self) -> None:
        """Test getting a profile that doesn't exist."""
        loader = ProfileLoader()
        with pytest.raises(KeyError, match="Profile not found"):
            loader.get("nonexistent-profile-xyz")

    def test_profile_caching(self) -> None:
        """Test that profiles are cached."""
        loader = ProfileLoader()
        profile1 = loader.get("base")
        profile2 = loader.get("base")
        assert profile1 is profile2  # Same object due to caching

    def test_resolve_dependencies_simple(self) -> None:
        """Test resolving simple dependencies."""
        loader = ProfileLoader()
        profiles = loader.resolve_dependencies(["tailscale"])
        names = [p.name for p in profiles]
        # base should come before tailscale
        assert names.index("base") < names.index("tailscale")

    def test_resolve_dependencies_chain(self) -> None:
        """Test resolving chained dependencies."""
        loader = ProfileLoader()
        # sunshine depends on gui, which depends on base
        profiles = loader.resolve_dependencies(["sunshine"])
        names = [p.name for p in profiles]
        assert names.index("base") < names.index("gui")
        assert names.index("gui") < names.index("sunshine")

    def test_resolve_dependencies_no_duplicates(self) -> None:
        """Test that dependencies aren't duplicated."""
        loader = ProfileLoader()
        profiles = loader.resolve_dependencies(["tailscale", "dev-python"])
        names = [p.name for p in profiles]
        # base should only appear once
        assert names.count("base") == 1


class TestGlobalLoader:
    """Tests for global profile loader."""

    def test_get_profile_loader(self) -> None:
        """Test getting global loader instance."""
        loader1 = get_profile_loader()
        loader2 = get_profile_loader()
        # Should be the same instance
        assert loader1 is loader2


class TestMergedProfile:
    """Tests for MergedProfile class."""

    def test_empty_merged_profile(self) -> None:
        """Test creating an empty merged profile."""
        merged = MergedProfile()
        assert merged.profiles == []
        assert merged.packages == []
        assert merged.services == []
        assert merged.requires_privileged is False

    def test_merge_single_profile(self) -> None:
        """Test merging a single profile."""
        profile = Profile(
            name="test",
            packages=["pkg1", "pkg2"],
            services=["svc1"],
            environment={"KEY": "value"},
        )
        merged = MergedProfile()
        merged.merge_profile(profile)

        assert merged.profiles == ["test"]
        assert merged.packages == ["pkg1", "pkg2"]
        assert merged.services == ["svc1"]
        assert merged.environment == {"KEY": "value"}

    def test_merge_multiple_profiles(self) -> None:
        """Test merging multiple profiles."""
        profile1 = Profile(
            name="profile1",
            packages=["pkg1", "common"],
            environment={"KEY1": "val1", "SHARED": "from1"},
        )
        profile2 = Profile(
            name="profile2",
            packages=["pkg2", "common"],
            environment={"KEY2": "val2", "SHARED": "from2"},
        )
        merged = MergedProfile()
        merged.merge_profile(profile1)
        merged.merge_profile(profile2)

        assert merged.profiles == ["profile1", "profile2"]
        # Packages should be deduplicated
        assert merged.packages == ["pkg1", "common", "pkg2"]
        # Environment from later profile overrides
        assert merged.environment["SHARED"] == "from2"
        assert merged.environment["KEY1"] == "val1"
        assert merged.environment["KEY2"] == "val2"

    def test_merge_capabilities(self) -> None:
        """Test merging capabilities from profiles."""
        profile = Profile(
            name="test",
            capabilities_add=["NET_ADMIN", "SYS_PTRACE"],
            capabilities_drop=["NET_RAW"],
        )
        merged = MergedProfile()
        merged.merge_profile(profile)

        assert "NET_ADMIN" in merged.capabilities_add
        assert "SYS_PTRACE" in merged.capabilities_add
        assert "NET_RAW" in merged.capabilities_drop

    def test_merge_privileged_flag(self) -> None:
        """Test that privileged flag is set if any profile requires it."""
        profile1 = Profile(name="normal", requires_privileged=False)
        profile2 = Profile(name="privileged", requires_privileged=True)
        profile3 = Profile(name="normal2", requires_privileged=False)

        merged = MergedProfile()
        merged.merge_profile(profile1)
        assert merged.requires_privileged is False

        merged.merge_profile(profile2)
        assert merged.requires_privileged is True

        merged.merge_profile(profile3)
        # Should still be True
        assert merged.requires_privileged is True

    def test_generate_setup_script_debian(self) -> None:
        """Test generating setup script for Debian."""
        merged = MergedProfile(
            packages=["vim", "curl"],
            services=["nginx"],
            setup_commands=["echo 'Hello World'"],
        )
        script = merged.generate_setup_script("trixie")

        assert "#!/bin/bash" in script
        assert "apt-get update" in script
        assert "apt-get install -y vim curl" in script
        assert "echo 'Hello World'" in script
        assert "systemctl enable nginx" in script

    def test_generate_setup_script_alpine(self) -> None:
        """Test generating setup script for Alpine."""
        merged = MergedProfile(packages=["vim"])
        script = merged.generate_setup_script("alpine")

        assert "apk update" in script
        assert "apk add --no-cache vim" in script

    def test_generate_setup_script_fedora(self) -> None:
        """Test generating setup script for Fedora."""
        merged = MergedProfile(packages=["vim"])
        script = merged.generate_setup_script("fedora")

        assert "dnf" in script
        # vim is translated to vim-enhanced on Fedora
        assert "vim-enhanced" in script

    def test_generate_setup_script_with_files(self) -> None:
        """Test generating setup script with file creation."""
        merged = MergedProfile(
            files={"/etc/test.conf": "key=value\n"}
        )
        script = merged.generate_setup_script("trixie")

        assert "cat > '/etc/test.conf'" in script
        assert "key=value" in script
        assert "GAOL_EOF" in script


class TestProfileApplicator:
    """Tests for ProfileApplicator class."""

    def test_resolve_and_merge_empty(self) -> None:
        """Test resolving empty profile list."""
        applicator = ProfileApplicator()
        merged = applicator.resolve_and_merge([])
        assert merged.profiles == []

    def test_resolve_and_merge_single(self) -> None:
        """Test resolving single profile."""
        applicator = ProfileApplicator()
        merged = applicator.resolve_and_merge(["base"])
        assert "base" in merged.profiles

    def test_resolve_and_merge_with_dependencies(self) -> None:
        """Test resolving profiles with dependencies."""
        applicator = ProfileApplicator()
        merged = applicator.resolve_and_merge(["tailscale"])
        # Should include base (dependency) and tailscale
        assert "base" in merged.profiles
        assert "tailscale" in merged.profiles
        # base should come before tailscale
        assert merged.profiles.index("base") < merged.profiles.index("tailscale")

    def test_apply_to_config_environment(self) -> None:
        """Test applying profile environment to config."""
        config = ContainerConfig(
            name="test",
            distro="trixie",
            environment={"USER_VAR": "user_value"},
        )
        merged = MergedProfile(
            profiles=["test"],
            environment={"PROFILE_VAR": "profile_value", "USER_VAR": "profile_default"},
        )
        applicator = ProfileApplicator()
        updated = applicator.apply_to_config(config, merged)

        # User value should override profile default
        assert updated.environment["USER_VAR"] == "user_value"
        # Profile var should be present
        assert updated.environment["PROFILE_VAR"] == "profile_value"

    def test_apply_to_config_volumes(self) -> None:
        """Test applying profile volumes to config."""
        config = ContainerConfig(
            name="test",
            distro="trixie",
            volumes={"/host/user": "/container/user"},
        )
        merged = MergedProfile(
            profiles=["test"],
            volumes={"/host/profile": "/container/profile", "/host/user": "/container/default"},
        )
        applicator = ProfileApplicator()
        updated = applicator.apply_to_config(config, merged)

        # User volume should override
        assert updated.volumes["/host/user"] == "/container/user"
        # Profile volume should be present
        assert updated.volumes["/host/profile"] == "/container/profile"

    def test_apply_to_config_ports(self) -> None:
        """Test applying profile exposed ports to config."""
        config = ContainerConfig(
            name="test",
            distro="trixie",
            network=NetworkConfig(ports={8080: 80}),
        )
        merged = MergedProfile(
            profiles=["test"],
            exposed_ports=[443, 80],  # 80 already mapped
        )
        applicator = ProfileApplicator()
        updated = applicator.apply_to_config(config, merged)

        # Existing mapping preserved
        assert updated.network.ports[8080] == 80
        # New port added (443 -> 443)
        assert 443 in updated.network.ports

    def test_apply_to_config_privileged(self) -> None:
        """Test applying privileged flag from profile."""
        config = ContainerConfig(
            name="test",
            distro="trixie",
            privileged=False,
        )
        merged = MergedProfile(
            profiles=["test"],
            requires_privileged=True,
        )
        applicator = ProfileApplicator()
        updated = applicator.apply_to_config(config, merged)

        assert updated.privileged is True

    def test_apply_to_config_preserves_profiles(self) -> None:
        """Test that applied profiles are stored in config."""
        config = ContainerConfig(name="test", distro="trixie")
        merged = MergedProfile(profiles=["base", "tailscale"])
        applicator = ProfileApplicator()
        updated = applicator.apply_to_config(config, merged)

        assert updated.profiles == ["base", "tailscale"]


class TestGetProfileApplicator:
    """Tests for get_profile_applicator function."""

    def test_get_profile_applicator(self) -> None:
        """Test getting profile applicator."""
        applicator = get_profile_applicator()
        assert isinstance(applicator, ProfileApplicator)



"""btrfs backend for Copy-on-Write filesystem support."""

from __future__ import annotations

import os
import shutil
import subprocess
from datetime import datetime
from pathlib import Path

from gaol.cow.backends.base import CoWBackendBase
from gaol.cow.exceptions import (
    BackendNotAvailableError,
    CoWError,
    DatasetExistsError,
    InvalidMountPointError,
    SendReceiveError,
    SnapshotExistsError,
    SnapshotHasClonesError,
)
from gaol.cow.models import (
    CoWBackendType,
    DatasetConfig,
    DatasetState,
    PoolConfig,
    PoolState,
    PoolStatus,
    Snapshot,
)


def _run_btrfs(args: list[str], check: bool = True) -> subprocess.CompletedProcess[str]:
    """Run a btrfs command."""
    cmd = ["btrfs"] + args
    return subprocess.run(
        cmd,
        capture_output=True,
        text=True,
        check=check,
    )


class BtrfsBackend(CoWBackendBase):
    """btrfs subvolume-based CoW backend."""

    @property
    def name(self) -> str:
        return "btrfs"

    def is_available(self) -> bool:
        """Check if btrfs tools are available."""
        return shutil.which("btrfs") is not None

    def _check_available(self) -> None:
        """Raise if btrfs is not available."""
        if not self.is_available():
            raise BackendNotAvailableError("btrfs", "btrfs-progs not installed")

    def validate_mount_point(self, path: Path) -> bool:
        """Check if path is a btrfs filesystem."""
        self._check_available()

        if not path.exists() or not path.is_dir():
            return False

        try:
            result = _run_btrfs(["filesystem", "show", str(path)], check=False)
            return result.returncode == 0
        except Exception:
            return False

    def _get_subvolume_path(self, pool: PoolState, name: str | None = None) -> Path:
        """Get the path to a subvolume."""
        base = pool.gaol_path
        if name:
            return base / name
        return base

    def _get_snapshot_path(self, pool: PoolState, dataset: str, snapshot: str) -> Path:
        """Get the path to a snapshot."""
        return pool.gaol_path / ".snapshots" / dataset / snapshot

    def _get_filesystem_info(self, path: Path) -> dict[str, int]:
        """Get filesystem usage information."""
        try:
            stat = os.statvfs(path)
            total = stat.f_blocks * stat.f_frsize
            free = stat.f_bavail * stat.f_frsize
            used = total - free
            return {"total": total, "used": used, "free": free}
        except Exception:
            return {"total": 0, "used": 0, "free": 0}

    def _get_subvolume_usage(self, path: Path) -> dict[str, int]:
        """Get subvolume usage via btrfs qgroup (if available)."""
        try:
            result = _run_btrfs(["qgroup", "show", "-r", str(path)], check=False)
            if result.returncode == 0:
                # Parse qgroup output - format varies
                lines = result.stdout.strip().split("\n")
                for line in lines[2:]:  # Skip headers
                    parts = line.split()
                    if len(parts) >= 3:
                        return {
                            "used": int(parts[1]),
                            "referenced": int(parts[2]) if len(parts) > 2 else 0,
                        }
        except Exception:
            pass
        return {"used": 0, "referenced": 0}

    # Pool operations

    def init_pool(self, config: PoolConfig) -> PoolState:
        """Initialize a btrfs pool."""
        self._check_available()

        if not self.validate_mount_point(config.mount_point):
            raise InvalidMountPointError(str(config.mount_point), "btrfs")

        # Create gaol subvolume
        gaol_path = config.mount_point / "gaol"

        if gaol_path.exists():
            # Check if it's already a subvolume
            result = _run_btrfs(["subvolume", "show", str(gaol_path)], check=False)
            if result.returncode != 0:
                raise CoWError(
                    f"Path {gaol_path} exists but is not a btrfs subvolume"
                )
        else:
            try:
                _run_btrfs(["subvolume", "create", str(gaol_path)])
            except subprocess.CalledProcessError as e:
                raise CoWError(f"Failed to create subvolume: {e.stderr}") from e

        # Create snapshots directory
        snapshots_dir = gaol_path / ".snapshots"
        snapshots_dir.mkdir(exist_ok=True)

        # Get filesystem info
        fs_info = self._get_filesystem_info(config.mount_point)

        return PoolState(
            name=config.name,
            backend=CoWBackendType.BTRFS,
            status=PoolStatus.ONLINE,
            mount_point=config.mount_point,
            gaol_path=gaol_path,
            total_bytes=fs_info["total"],
            used_bytes=fs_info["used"],
            free_bytes=fs_info["free"],
            compression=config.compression,
            created_at=datetime.now(),
        )

    def get_pool_info(self, state: PoolState) -> PoolState:
        """Get updated pool information."""
        self._check_available()

        fs_info = self._get_filesystem_info(state.mount_point)

        # Check if subvolume still exists
        status = PoolStatus.ONLINE
        if not state.gaol_path.exists():
            status = PoolStatus.ERROR

        return PoolState(
            name=state.name,
            backend=state.backend,
            status=status,
            mount_point=state.mount_point,
            gaol_path=state.gaol_path,
            total_bytes=fs_info["total"],
            used_bytes=fs_info["used"],
            free_bytes=fs_info["free"],
            compression=state.compression,
            created_at=state.created_at,
        )

    def destroy_pool(self, state: PoolState) -> None:
        """Destroy a btrfs pool."""
        self._check_available()

        if not state.gaol_path.exists():
            return

        # Delete all subvolumes recursively
        self._delete_subvolume_recursive(state.gaol_path)

    def _delete_subvolume_recursive(self, path: Path) -> None:
        """Recursively delete a subvolume and all nested subvolumes."""
        if not path.exists():
            return

        # List child subvolumes
        try:
            result = _run_btrfs(
                ["subvolume", "list", "-o", str(path)], check=False
            )
            if result.returncode == 0:
                for line in result.stdout.strip().split("\n"):
                    if line:
                        # Parse "ID xxx gen xxx top level xxx path <path>"
                        parts = line.split("path ")
                        if len(parts) > 1:
                            child_rel_path = parts[1]
                            # Reconstruct full path
                            child_path = path.parent / child_rel_path
                            if child_path.exists():
                                self._delete_subvolume_recursive(child_path)
        except Exception:
            pass

        # Delete this subvolume
        try:
            _run_btrfs(["subvolume", "delete", str(path)])
        except subprocess.CalledProcessError:
            # Try with force
            try:
                shutil.rmtree(path, ignore_errors=True)
            except Exception:
                pass

    # Dataset operations

    def create_dataset(self, pool: PoolState, config: DatasetConfig) -> DatasetState:
        """Create a new btrfs subvolume."""
        self._check_available()

        subvol_path = self._get_subvolume_path(pool, config.name)

        if subvol_path.exists():
            raise DatasetExistsError(pool.name, config.name)

        try:
            _run_btrfs(["subvolume", "create", str(subvol_path)])
        except subprocess.CalledProcessError as e:
            raise CoWError(f"Failed to create subvolume: {e.stderr}") from e

        # Create snapshot dir for this dataset
        snapshot_dir = pool.gaol_path / ".snapshots" / config.name
        snapshot_dir.mkdir(parents=True, exist_ok=True)

        # Set quota if requested
        if config.quota_mb:
            self._set_quota_internal(subvol_path, config.quota_mb)

        usage = self._get_subvolume_usage(subvol_path)

        return DatasetState(
            name=config.name,
            pool=pool.name,
            path=subvol_path,
            quota_mb=config.quota_mb,
            used_bytes=usage["used"],
            referenced_bytes=usage.get("referenced", 0),
            compression=config.compression or pool.compression,
            created_at=datetime.now(),
            snapshots=[],
        )

    def get_dataset_info(self, pool: PoolState, state: DatasetState) -> DatasetState:
        """Get updated dataset information."""
        self._check_available()

        usage = self._get_subvolume_usage(state.path)

        # List snapshots
        snapshot_dir = pool.gaol_path / ".snapshots" / state.name
        snapshots = []
        if snapshot_dir.exists():
            snapshots = [d.name for d in snapshot_dir.iterdir() if d.is_dir()]

        return DatasetState(
            name=state.name,
            pool=state.pool,
            path=state.path,
            quota_mb=state.quota_mb,
            used_bytes=usage["used"],
            referenced_bytes=usage.get("referenced", 0),
            compression=state.compression,
            created_at=state.created_at,
            snapshots=snapshots,
        )

    def destroy_dataset(self, pool: PoolState, state: DatasetState) -> None:
        """Destroy a dataset and its snapshots."""
        self._check_available()

        # Delete snapshots first
        snapshot_dir = pool.gaol_path / ".snapshots" / state.name
        if snapshot_dir.exists():
            for snap in snapshot_dir.iterdir():
                if snap.is_dir():
                    self._delete_subvolume_recursive(snap)
            snapshot_dir.rmdir()

        # Delete the dataset subvolume
        if state.path.exists():
            self._delete_subvolume_recursive(state.path)

    def _set_quota_internal(self, path: Path, quota_mb: int | None) -> None:
        """Set or remove quota on a subvolume."""
        try:
            # Enable quotas first
            _run_btrfs(["quota", "enable", str(path.parent)], check=False)

            if quota_mb:
                quota_bytes = quota_mb * 1024 * 1024
                # Get subvolume ID
                result = _run_btrfs(["subvolume", "show", str(path)])
                for line in result.stdout.split("\n"):
                    if "Subvolume ID:" in line:
                        subvol_id = line.split()[-1]
                        _run_btrfs([
                            "qgroup", "limit", str(quota_bytes),
                            f"0/{subvol_id}", str(path)
                        ], check=False)
                        break
            else:
                # Remove quota by setting to 'none'
                result = _run_btrfs(["subvolume", "show", str(path)])
                for line in result.stdout.split("\n"):
                    if "Subvolume ID:" in line:
                        subvol_id = line.split()[-1]
                        _run_btrfs([
                            "qgroup", "limit", "none",
                            f"0/{subvol_id}", str(path)
                        ], check=False)
                        break
        except Exception:
            pass  # Quota might not be supported

    def set_quota(self, pool: PoolState, state: DatasetState, quota_mb: int | None) -> DatasetState:
        """Set or remove quota on a dataset."""
        self._check_available()
        self._set_quota_internal(state.path, quota_mb)

        return DatasetState(
            name=state.name,
            pool=state.pool,
            path=state.path,
            quota_mb=quota_mb,
            used_bytes=state.used_bytes,
            referenced_bytes=state.referenced_bytes,
            compression=state.compression,
            created_at=state.created_at,
            snapshots=state.snapshots,
        )

    # Snapshot operations

    def create_snapshot(
        self, pool: PoolState, dataset: DatasetState, name: str
    ) -> Snapshot:
        """Create a readonly snapshot."""
        self._check_available()

        snapshot_path = self._get_snapshot_path(pool, dataset.name, name)

        if snapshot_path.exists():
            raise SnapshotExistsError(pool.name, dataset.name, name)

        snapshot_path.parent.mkdir(parents=True, exist_ok=True)

        try:
            _run_btrfs([
                "subvolume", "snapshot", "-r",
                str(dataset.path), str(snapshot_path)
            ])
        except subprocess.CalledProcessError as e:
            raise CoWError(f"Failed to create snapshot: {e.stderr}") from e

        usage = self._get_subvolume_usage(snapshot_path)

        return Snapshot(
            name=name,
            dataset=dataset.name,
            pool=pool.name,
            created_at=datetime.now(),
            used_bytes=usage["used"],
            referenced_bytes=usage.get("referenced", 0),
            readonly=True,
        )

    def list_snapshots(
        self, pool: PoolState, dataset: DatasetState
    ) -> list[Snapshot]:
        """List all snapshots of a dataset."""
        self._check_available()

        snapshot_dir = pool.gaol_path / ".snapshots" / dataset.name
        snapshots = []

        if not snapshot_dir.exists():
            return snapshots

        for snap_path in sorted(snapshot_dir.iterdir()):
            if snap_path.is_dir():
                usage = self._get_subvolume_usage(snap_path)
                # Get creation time from path stat
                try:
                    created = datetime.fromtimestamp(snap_path.stat().st_ctime)
                except Exception:
                    created = datetime.now()

                snapshots.append(Snapshot(
                    name=snap_path.name,
                    dataset=dataset.name,
                    pool=pool.name,
                    created_at=created,
                    used_bytes=usage["used"],
                    referenced_bytes=usage.get("referenced", 0),
                    readonly=True,
                ))

        return snapshots

    def delete_snapshot(
        self, pool: PoolState, dataset: DatasetState, snapshot: Snapshot
    ) -> None:
        """Delete a snapshot."""
        self._check_available()

        snapshot_path = self._get_snapshot_path(pool, dataset.name, snapshot.name)

        if not snapshot_path.exists():
            return

        # Check for clones (btrfs doesn't have direct clone tracking,
        # so we skip this check for btrfs)

        try:
            _run_btrfs(["subvolume", "delete", str(snapshot_path)])
        except subprocess.CalledProcessError as e:
            raise CoWError(f"Failed to delete snapshot: {e.stderr}") from e

    def rollback(
        self, pool: PoolState, dataset: DatasetState, snapshot: Snapshot
    ) -> DatasetState:
        """Rollback a dataset to a snapshot."""
        self._check_available()

        snapshot_path = self._get_snapshot_path(pool, dataset.name, snapshot.name)

        if not snapshot_path.exists():
            raise CoWError(f"Snapshot not found: {snapshot_path}")

        # Rename current subvolume
        backup_path = dataset.path.with_suffix(".rollback-backup")
        if backup_path.exists():
            self._delete_subvolume_recursive(backup_path)

        try:
            # Rename current to backup
            dataset.path.rename(backup_path)

            # Create writable snapshot from the readonly snapshot
            _run_btrfs([
                "subvolume", "snapshot",
                str(snapshot_path), str(dataset.path)
            ])

            # Delete backup
            self._delete_subvolume_recursive(backup_path)

        except Exception as e:
            # Restore backup on failure
            if backup_path.exists() and not dataset.path.exists():
                backup_path.rename(dataset.path)
            raise CoWError(f"Rollback failed: {e}") from e

        return self.get_dataset_info(pool, dataset)

    # Clone operations

    def create_clone(
        self, pool: PoolState, snapshot: Snapshot, name: str
    ) -> DatasetState:
        """Create a writable clone from a snapshot."""
        self._check_available()

        snapshot_path = self._get_snapshot_path(pool, snapshot.dataset, snapshot.name)
        clone_path = self._get_subvolume_path(pool, name)

        if clone_path.exists():
            raise DatasetExistsError(pool.name, name)

        try:
            # Create writable snapshot (clone)
            _run_btrfs([
                "subvolume", "snapshot",
                str(snapshot_path), str(clone_path)
            ])
        except subprocess.CalledProcessError as e:
            raise CoWError(f"Failed to create clone: {e.stderr}") from e

        # Create snapshot dir for the clone
        snapshot_dir = pool.gaol_path / ".snapshots" / name
        snapshot_dir.mkdir(parents=True, exist_ok=True)

        usage = self._get_subvolume_usage(clone_path)

        return DatasetState(
            name=name,
            pool=pool.name,
            path=clone_path,
            quota_mb=None,
            used_bytes=usage["used"],
            referenced_bytes=usage.get("referenced", 0),
            compression=pool.compression,
            created_at=datetime.now(),
            snapshots=[],
        )

    def promote_clone(
        self, pool: PoolState, clone: DatasetState
    ) -> DatasetState:
        """Promote a clone to independent dataset.

        btrfs clones are already independent (no dependency tracking),
        so this is a no-op.
        """
        return clone

    # Send/Receive operations

    def send(
        self,
        pool: PoolState,
        dataset: DatasetState,
        snapshot: Snapshot,
        dest: Path,
        base_snapshot: Snapshot | None = None,
    ) -> Path:
        """Send a snapshot to a file."""
        self._check_available()

        snapshot_path = self._get_snapshot_path(pool, dataset.name, snapshot.name)

        if not snapshot_path.exists():
            raise SendReceiveError("send", f"Snapshot not found: {snapshot_path}")

        try:
            cmd = ["btrfs", "send"]

            if base_snapshot:
                base_path = self._get_snapshot_path(
                    pool, dataset.name, base_snapshot.name
                )
                cmd.extend(["-p", str(base_path)])

            cmd.append(str(snapshot_path))

            # Pipe to file
            with open(dest, "wb") as f:
                result = subprocess.run(
                    cmd,
                    stdout=f,
                    stderr=subprocess.PIPE,
                    check=True,
                )

            return dest

        except subprocess.CalledProcessError as e:
            raise SendReceiveError("send", e.stderr.decode()) from e

    def receive(
        self,
        pool: PoolState,
        source: Path,
        name: str | None = None,
    ) -> DatasetState:
        """Receive a snapshot stream into a new dataset."""
        self._check_available()

        if not source.exists():
            raise SendReceiveError("receive", f"Source file not found: {source}")

        # Receive into gaol path
        dest_dir = pool.gaol_path

        try:
            with open(source, "rb") as f:
                result = subprocess.run(
                    ["btrfs", "receive", str(dest_dir)],
                    stdin=f,
                    capture_output=True,
                    check=True,
                )

            # Parse output to find received subvolume name
            # Output format: "At subvol <name>"
            output = result.stdout.decode()
            received_name = name
            for line in output.split("\n"):
                if "At subvol" in line or "At snapshot" in line:
                    received_name = line.split()[-1]
                    break

            if not received_name:
                received_name = source.stem

            # Find the received subvolume
            received_path = dest_dir / received_name

            if not received_path.exists():
                raise SendReceiveError(
                    "receive", f"Received subvolume not found: {received_path}"
                )

            # Rename if different name requested
            final_path = received_path
            if name and name != received_name:
                final_path = dest_dir / name
                received_path.rename(final_path)
                received_name = name

            # Create snapshot dir
            snapshot_dir = pool.gaol_path / ".snapshots" / received_name
            snapshot_dir.mkdir(parents=True, exist_ok=True)

            usage = self._get_subvolume_usage(final_path)

            return DatasetState(
                name=received_name,
                pool=pool.name,
                path=final_path,
                quota_mb=None,
                used_bytes=usage["used"],
                referenced_bytes=usage.get("referenced", 0),
                compression=pool.compression,
                created_at=datetime.now(),
                snapshots=[],
            )

        except subprocess.CalledProcessError as e:
            raise SendReceiveError("receive", e.stderr.decode()) from e



"""Profile models for container configuration."""

from typing import Annotated

from pydantic import BaseModel, Field


class ProfileDependency(BaseModel):
    """A dependency on another profile."""

    name: str
    optional: bool = False


class Profile(BaseModel):
    """A profile defines software and configuration for a container."""

    name: Annotated[str, Field(min_length=1, max_length=64)]
    description: str = ""
    version: str = "1.0.0"

    # Dependencies
    depends_on: Annotated[
        list[ProfileDependency], Field(default_factory=list, description="Required profiles")
    ]
    conflicts_with: Annotated[
        list[str], Field(default_factory=list, description="Incompatible profiles")
    ]

    # System requirements
    requires_root: bool = False
    requires_privileged: bool = False
    supported_distros: Annotated[
        list[str], Field(default_factory=list, description="Empty means all distros supported")
    ]

    # Packages to install
    packages: Annotated[list[str], Field(default_factory=list)]
    package_repos: Annotated[
        list[str], Field(default_factory=list, description="Additional apt/dnf repositories")
    ]

    # Services to enable/start
    services: Annotated[list[str], Field(default_factory=list)]

    # Commands to run during setup
    setup_commands: Annotated[
        list[str], Field(default_factory=list, description="Commands run after package install")
    ]

    # Files to create or modify
    files: Annotated[
        dict[str, str],
        Field(default_factory=dict, description="Path -> content for files to create"),
    ]

    # Environment variables
    environment: Annotated[
        dict[str, str],
        Field(default_factory=dict, description="Environment variables to set"),
    ]

    # Network requirements
    exposed_ports: Annotated[list[int], Field(default_factory=list)]

    # Volume mounts required by this profile
    volumes: Annotated[
        dict[str, str],
        Field(default_factory=dict, description="Host path -> container path"),
    ]

    # Capabilities and security
    capabilities_add: Annotated[list[str], Field(default_factory=list)]
    capabilities_drop: Annotated[list[str], Field(default_factory=list)]

    @classmethod
    def from_yaml(cls, yaml_content: dict) -> "Profile":
        """Create a Profile from parsed YAML content."""
        # Make a copy to avoid mutating the input
        data = yaml_content.copy()

        # Convert simple dependency strings to ProfileDependency objects
        if "depends_on" in data:
            deps = []
            for dep in data["depends_on"]:
                if isinstance(dep, str):
                    deps.append(ProfileDependency(name=dep))
                elif isinstance(dep, ProfileDependency):
                    deps.append(dep)
                elif isinstance(dep, dict):
                    deps.append(ProfileDependency(**dep))
                else:
                    raise TypeError(f"Unsupported dependency type: {type(dep)}")
            data["depends_on"] = deps
        return cls(**data)



"""Built-in templates for the template gallery.

These templates are bundled with gaol and available
without network access.
"""

from __future__ import annotations

from typing import Any

# Built-in templates as dictionaries
# These are converted to Template objects by the loader
BUILTIN_TEMPLATES: dict[str, dict[str, Any]] = {
    # ==========================================================================
    # Web Development
    # ==========================================================================
    "web-dev/django": {
        "metadata": {
            "id": "web-dev/django",
            "display_name": "Django Development",
            "description": "Full Django development environment with PostgreSQL support",
            "category": "web-dev",
            "tags": ["python", "django", "web", "postgresql", "framework"],
            "author": "gaol",
            "version": "1.0.0",
            "icon": "🐍",
        },
        "spec": {
            "name": "django-dev",
            "runtime": "docker",
            "distro": "trixie",
            "profiles": ["base", "dev-python"],
            "resources": {
                "memory": "2G",
                "cpus": 2,
            },
            "network": {
                "ports": ["8000:8000"],
            },
            "environment": {
                "DJANGO_DEBUG": "true",
                "DJANGO_SETTINGS_MODULE": "config.settings.local",
            },
            "volumes": ["~/projects:/workspace"],
            "working_dir": "/workspace",
        },
        "post_create_commands": [
            "uv pip install django psycopg2-binary django-debug-toolbar",
        ],
        "example_env": {
            "DATABASE_URL": "postgres://user:pass@localhost:5432/mydb",
            "SECRET_KEY": "your-secret-key-here",
            "DJANGO_DEBUG": "true",
        },
    },
    "web-dev/flask": {
        "metadata": {
            "id": "web-dev/flask",
            "display_name": "Flask Development",
            "description": "Lightweight Flask development environment",
            "category": "web-dev",
            "tags": ["python", "flask", "web", "api", "microframework"],
            "author": "gaol",
            "version": "1.0.0",
            "icon": "🧪",
        },
        "spec": {
            "name": "flask-dev",
            "runtime": "docker",
            "distro": "trixie",
            "profiles": ["base", "dev-python"],
            "resources": {
                "memory": "1G",
                "cpus": 1,
            },
            "network": {
                "ports": ["5000:5000"],
            },
            "environment": {
                "FLASK_DEBUG": "1",
                "FLASK_APP": "app.py",
            },
            "volumes": ["~/projects:/workspace"],
            "working_dir": "/workspace",
        },
        "post_create_commands": [
            "uv pip install flask flask-sqlalchemy flask-migrate",
        ],
    },
    "web-dev/fastapi": {
        "metadata": {
            "id": "web-dev/fastapi",
            "display_name": "FastAPI Development",
            "description": "Modern async API development with FastAPI",
            "category": "web-dev",
            "tags": ["python", "fastapi", "api", "async", "openapi"],
            "author": "gaol",
            "version": "1.0.0",
            "icon": "⚡",
        },
        "spec": {
            "name": "fastapi-dev",
            "runtime": "docker",
            "distro": "trixie",
            "profiles": ["base", "dev-python"],
            "resources": {
                "memory": "2G",
                "cpus": 2,
            },
            "network": {
                "ports": ["8000:8000"],
            },
            "environment": {
                "PYTHONPATH": "/workspace",
            },
            "volumes": ["~/projects:/workspace"],
            "working_dir": "/workspace",
        },
        "post_create_commands": [
            "uv pip install 'fastapi[standard]' sqlalchemy alembic",
        ],
    },
    "web-dev/node": {
        "metadata": {
            "id": "web-dev/node",
            "display_name": "Node.js Development",
            "description": "Node.js development with pnpm and TypeScript",
            "category": "web-dev",
            "tags": ["node", "javascript", "typescript", "pnpm", "web"],
            "author": "gaol",
            "version": "1.0.0",
            "icon": "🟢",
        },
        "spec": {
            "name": "node-dev",
            "runtime": "docker",
            "distro": "trixie",
            "profiles": ["base", "dev-node"],
            "resources": {
                "memory": "2G",
                "cpus": 2,
            },
            "network": {
                "ports": ["3000:3000", "5173:5173"],
            },
            "volumes": ["~/projects:/workspace"],
            "working_dir": "/workspace",
        },
        "post_create_commands": [
            "pnpm install -g typescript ts-node @types/node",
        ],
    },
    "web-dev/rails": {
        "metadata": {
            "id": "web-dev/rails",
            "display_name": "Ruby on Rails",
            "description": "Ruby on Rails development environment",
            "category": "web-dev",
            "tags": ["ruby", "rails", "web", "postgresql", "framework"],
            "author": "gaol",
            "version": "1.0.0",
            "icon": "💎",
        },
        "spec": {
            "name": "rails-dev",
            "runtime": "docker",
            "distro": "trixie",
            "profiles": ["base"],
            "resources": {
                "memory": "2G",
                "cpus": 2,
            },
            "network": {
                "ports": ["3000:3000"],
            },
            "environment": {
                "RAILS_ENV": "development",
            },
            "volumes": ["~/projects:/workspace"],
            "working_dir": "/workspace",
        },
        "post_create_commands": [
            "apt-get update && apt-get install -y ruby ruby-dev build-essential libpq-dev",
            "gem install rails bundler",
        ],
    },
    # ==========================================================================
    # Data Science
    # ==========================================================================
    "data-science/jupyter": {
        "metadata": {
            "id": "data-science/jupyter",
            "display_name": "Jupyter Lab",
            "description": "Jupyter Lab with data science packages",
            "category": "data-science",
            "tags": ["python", "jupyter", "pandas", "numpy", "matplotlib"],
            "author": "gaol",
            "version": "1.0.0",
            "icon": "📊",
        },
        "spec": {
            "name": "jupyter-lab",
            "runtime": "docker",
            "distro": "trixie",
            "profiles": ["base", "dev-python"],
            "resources": {
                "memory": "4G",
                "cpus": 2,
            },
            "network": {
                "ports": ["8888:8888"],
            },
            "volumes": ["~/notebooks:/workspace"],
            "working_dir": "/workspace",
        },
        "post_create_commands": [
            "uv pip install jupyterlab pandas numpy matplotlib seaborn scikit-learn",
        ],
        "readme": """# Jupyter Lab Template

Start Jupyter Lab with:
```bash
gaol exec jupyter-lab -- jupyter lab --ip=0.0.0.0 --no-browser
```

Then open http://localhost:8888 in your browser.
""",
    },
    "data-science/rstudio": {
        "metadata": {
            "id": "data-science/rstudio",
            "display_name": "RStudio Server",
            "description": "RStudio Server for R development",
            "category": "data-science",
            "tags": ["r", "rstudio", "statistics", "data-analysis"],
            "author": "gaol",
            "version": "1.0.0",
            "icon": "📈",
        },
        "spec": {
            "name": "rstudio",
            "runtime": "docker",
            "distro": "trixie",
            "profiles": ["base"],
            "resources": {
                "memory": "4G",
                "cpus": 2,
            },
            "network": {
                "ports": ["8787:8787"],
            },
            "volumes": ["~/r-projects:/home/rstudio"],
        },
        "post_create_commands": [
            "apt-get update && apt-get install -y r-base r-base-dev",
            "wget https://download2.rstudio.org/server/jammy/amd64/rstudio-server-2023.12.1-402-amd64.deb",
            "dpkg -i rstudio-server-*.deb || apt-get install -f -y",
            "rm rstudio-server-*.deb",
        ],
    },
    "data-science/mlflow": {
        "metadata": {
            "id": "data-science/mlflow",
            "display_name": "MLflow",
            "description": "MLflow for ML experiment tracking",
            "category": "data-science",
            "tags": ["python", "mlflow", "ml", "experiment-tracking"],
            "author": "gaol",
            "version": "1.0.0",
            "icon": "🤖",
        },
        "spec": {
            "name": "mlflow",
            "runtime": "docker",
            "distro": "trixie",
            "profiles": ["base", "dev-python"],
            "resources": {
                "memory": "2G",
                "cpus": 2,
            },
            "network": {
                "ports": ["5000:5000"],
            },
            "volumes": ["~/mlruns:/mlruns"],
            "environment": {
                "MLFLOW_TRACKING_URI": "/mlruns",
            },
        },
        "post_create_commands": [
            "uv pip install mlflow scikit-learn pandas",
        ],
    },
    # ==========================================================================
    # DevOps
    # ==========================================================================
    "devops/kubernetes": {
        "metadata": {
            "id": "devops/kubernetes",
            "display_name": "Kubernetes Tools",
            "description": "kubectl, helm, and k9s for Kubernetes management",
            "category": "devops",
            "tags": ["kubernetes", "kubectl", "helm", "k9s", "devops"],
            "author": "gaol",
            "version": "1.0.0",
            "icon": "☸️",
        },
        "spec": {
            "name": "k8s-tools",
            "runtime": "docker",
            "distro": "trixie",
            "profiles": ["base"],
            "resources": {
                "memory": "1G",
                "cpus": 1,
            },
            "volumes": [
                "~/.kube:/root/.kube:ro",
                "~/k8s:/workspace",
            ],
            "working_dir": "/workspace",
        },
        "post_create_commands": [
            "curl -LO https://dl.k8s.io/release/$(curl -L -s https://dl.k8s.io/release/stable.txt)/bin/linux/amd64/kubectl && chmod +x kubectl && mv kubectl /usr/local/bin/",
            "curl https://raw.githubusercontent.com/helm/helm/main/scripts/get-helm-3 | bash",
            "curl -sS https://webinstall.dev/k9s | bash",
        ],
    },
    "devops/terraform": {
        "metadata": {
            "id": "devops/terraform",
            "display_name": "Terraform",
            "description": "Terraform for infrastructure as code",
            "category": "devops",
            "tags": ["terraform", "iac", "infrastructure", "devops"],
            "author": "gaol",
            "version": "1.0.0",
            "icon": "🏗️",
        },
        "spec": {
            "name": "terraform",
            "runtime": "docker",
            "distro": "trixie",
            "profiles": ["base"],
            "resources": {
                "memory": "1G",
                "cpus": 1,
            },
            "volumes": ["~/terraform:/workspace"],
            "working_dir": "/workspace",
        },
        "post_create_commands": [
            "apt-get update && apt-get install -y gnupg software-properties-common",
            "wget -O- https://apt.releases.hashicorp.com/gpg | gpg --dearmor -o /usr/share/keyrings/hashicorp-archive-keyring.gpg",
            "echo 'deb [signed-by=/usr/share/keyrings/hashicorp-archive-keyring.gpg] https://apt.releases.hashicorp.com bookworm main' > /etc/apt/sources.list.d/hashicorp.list",
            "apt-get update && apt-get install -y terraform",
        ],
    },
    # ==========================================================================
    # Databases
    # ==========================================================================
    "database/postgres": {
        "metadata": {
            "id": "database/postgres",
            "display_name": "PostgreSQL",
            "description": "PostgreSQL database server with pgAdmin",
            "category": "database",
            "tags": ["postgresql", "database", "sql"],
            "author": "gaol",
            "version": "1.0.0",
            "icon": "🐘",
        },
        "spec": {
            "name": "postgres",
            "runtime": "docker",
            "distro": "trixie",
            "profiles": ["base"],
            "resources": {
                "memory": "1G",
                "cpus": 1,
            },
            "network": {
                "ports": ["5432:5432"],
            },
            "environment": {
                "POSTGRES_USER": "postgres",
                "POSTGRES_DB": "postgres",
            },
            "volumes": ["~/pgdata:/var/lib/postgresql/data"],
        },
        "post_create_commands": [
            "apt-get update && apt-get install -y postgresql postgresql-contrib",
        ],
        "example_env": {
            "POSTGRES_PASSWORD": "your-secure-password",
        },
    },
    "database/redis": {
        "metadata": {
            "id": "database/redis",
            "display_name": "Redis",
            "description": "Redis in-memory data store",
            "category": "database",
            "tags": ["redis", "cache", "database", "key-value"],
            "author": "gaol",
            "version": "1.0.0",
            "icon": "🔴",
        },
        "spec": {
            "name": "redis",
            "runtime": "docker",
            "distro": "trixie",
            "profiles": ["base"],
            "resources": {
                "memory": "512M",
                "cpus": 1,
            },
            "network": {
                "ports": ["6379:6379"],
            },
            "volumes": ["~/redis-data:/data"],
        },
        "post_create_commands": [
            "apt-get update && apt-get install -y redis-server",
        ],
    },
    # ==========================================================================
    # Gaming
    # ==========================================================================
    "gaming/steam": {
        "metadata": {
            "id": "gaming/steam",
            "display_name": "Steam",
            "description": "Steam gaming client with GPU support",
            "category": "gaming",
            "tags": ["steam", "gaming", "gpu"],
            "author": "gaol",
            "version": "1.0.0",
            "icon": "🎮",
            "supported_runtimes": ["docker"],
        },
        "spec": {
            "name": "steam",
            "runtime": "docker",
            "distro": "trixie",
            "profiles": ["base", "gui"],
            "resources": {
                "memory": "8G",
                "cpus": 4,
            },
            "privileged": True,
        },
        "post_create_commands": [
            "dpkg --add-architecture i386",
            "apt-get update",
            "apt-get install -y steam-installer",
        ],
    },
    # ==========================================================================
    # Security
    # ==========================================================================
    "security/kali-tools": {
        "metadata": {
            "id": "security/kali-tools",
            "display_name": "Kali Tools",
            "description": "Common Kali Linux security tools",
            "category": "security",
            "tags": ["kali", "security", "pentesting", "tools"],
            "author": "gaol",
            "version": "1.0.0",
            "icon": "🔒",
        },
        "spec": {
            "name": "kali-tools",
            "runtime": "docker",
            "distro": "trixie",
            "profiles": ["base"],
            "resources": {
                "memory": "2G",
                "cpus": 2,
            },
            "network": {
                "mode": "host",
            },
            "privileged": True,
        },
        "post_create_commands": [
            "apt-get update && apt-get install -y nmap nikto sqlmap hydra john hashcat metasploit-framework",
        ],
    },
}


def get_builtin_template(template_id: str) -> dict[str, Any] | None:
    """Get a built-in template by ID.

    Args:
        template_id: Template ID (e.g., 'web-dev/django')

    Returns:
        Template dict or None if not found
    """
    return BUILTIN_TEMPLATES.get(template_id)


def list_builtin_templates() -> list[str]:
    """List all built-in template IDs.

    Returns:
        List of template IDs
    """
    return list(BUILTIN_TEMPLATES.keys())


def get_builtin_categories() -> set[str]:
    """Get all categories from built-in templates.

    Returns:
        Set of category names
    """
    categories = set()
    for template_id in BUILTIN_TEMPLATES:
        category = template_id.split("/")[0]
        categories.add(category)
    return categories



"""Remote host management for gaol.

This module provides remote host management for executing gaol
commands on remote hosts via SSH. Features include:
- Register and manage remote hosts
- Execute gaol commands on remote hosts
- Check host connectivity and gaol availability
- Sync container configurations between machines

Example:
    >>> from gaol.remote import get_remote_manager, HostConfig
    >>> manager = get_remote_manager()
    >>> config = HostConfig(
    ...     name="production",
    ...     hostname="prod.example.com",
    ...     user="deploy",
    ...     key_path="~/.ssh/prod_key"
    ... )
    >>> manager.add_host(config)
    >>> manager.check_host("production")
    >>> result = manager.exec_gaol("production", ["list"])
"""

from gaol.remote.exceptions import (
    ConfigSyncError,
    HostExistsError,
    HostNotFoundError,
    KeyNotFoundError,
    GaolNotFoundError,
    RemoteCommandError,
    RemoteError,
    SSHAuthenticationError,
    SSHConnectionError,
    SSHTimeoutError,
)
from gaol.remote.models import (
    HostConfig,
    HostConnectionStatus,
    HostsRegistry,
    HostState,
    RemoteCommandResult,
    StoredHost,
)
from gaol.remote.executor import SSHExecutor
from gaol.remote.manager import RemoteManager, get_remote_manager
from gaol.remote.store import HostStore

__all__ = [
    # Manager
    "RemoteManager",
    "get_remote_manager",
    # Executor
    "SSHExecutor",
    # Store
    "HostStore",
    # Models
    "HostConfig",
    "HostConnectionStatus",
    "HostsRegistry",
    "HostState",
    "RemoteCommandResult",
    "StoredHost",
    # Exceptions
    "RemoteError",
    "HostNotFoundError",
    "HostExistsError",
    "SSHConnectionError",
    "SSHAuthenticationError",
    "SSHTimeoutError",
    "RemoteCommandError",
    "GaolNotFoundError",
    "ConfigSyncError",
    "KeyNotFoundError",
]



"""GPU configuration models.

This module provides Pydantic models for GPU configuration and
device mapping used in container GPU passthrough.
"""

from __future__ import annotations

from enum import Enum

from pydantic import BaseModel, Field


class GPUVendor(str, Enum):
    """GPU vendor types."""

    NVIDIA = "nvidia"
    AMD = "amd"
    INTEL = "intel"
    AUTO = "auto"  # Auto-detect


class GPUInfo(BaseModel):
    """Information about a detected GPU.

    Attributes:
        index: GPU device index (0, 1, 2, etc.)
        name: Human-readable GPU name
        vendor: GPU vendor (nvidia, amd, intel)
        uuid: Unique identifier for the GPU
        memory_mb: Total GPU memory in megabytes
        pci_bus_id: PCI bus address for passthrough
        driver_version: Installed driver version
        cuda_version: CUDA version (NVIDIA only)
    """

    index: int
    name: str
    vendor: GPUVendor
    uuid: str | None = None
    memory_mb: int | None = None
    pci_bus_id: str | None = None
    driver_version: str | None = None
    cuda_version: str | None = None


class GPUConfig(BaseModel):
    """GPU configuration for containers.

    Attributes:
        enabled: Whether GPU passthrough is enabled
        vendor: GPU vendor to use (auto-detect if not specified)
        device_ids: List of GPU device IDs to pass through (empty = all)
        capabilities: NVIDIA capabilities (compute, utility, graphics, etc.)
        require_nvidia_runtime: Use nvidia-container-runtime for Docker
        memory_limit: Memory limit per GPU (e.g., "8g")

    Example:
        >>> config = GPUConfig(enabled=True, device_ids=[0, 1])
        >>> config.vendor
        <GPUVendor.AUTO: 'auto'>
    """

    enabled: bool = False
    vendor: GPUVendor = GPUVendor.AUTO
    device_ids: list[int] = Field(default_factory=list)
    capabilities: list[str] = Field(default_factory=lambda: ["compute", "utility"])
    require_nvidia_runtime: bool = True
    memory_limit: str | None = None


class DeviceMapping(BaseModel):
    """Generic device mapping for passthrough.

    Attributes:
        host_path: Device path on host (e.g., /dev/nvidia0)
        container_path: Device path in container (defaults to host_path)
        permissions: Device permissions (r=read, w=write, m=mknod)

    Example:
        >>> device = DeviceMapping(host_path="/dev/nvidia0")
        >>> device.container_path  # Defaults to host_path
        '/dev/nvidia0'
    """

    host_path: str
    container_path: str | None = None
    permissions: str = "rwm"

    @property
    def effective_container_path(self) -> str:
        """Get the effective container path (defaults to host_path)."""
        return self.container_path or self.host_path


class GPUSupport(BaseModel):
    """Information about GPU support on the system.

    Attributes:
        nvidia_available: Whether NVIDIA GPUs are detected
        nvidia_driver_version: NVIDIA driver version
        nvidia_runtime_available: Whether nvidia-container-runtime is installed
        amd_available: Whether AMD GPUs are detected
        rocm_available: Whether ROCm is installed
        intel_available: Whether Intel GPUs are detected
        gpus: List of detected GPUs
    """

    nvidia_available: bool = False
    nvidia_driver_version: str | None = None
    nvidia_runtime_available: bool = False
    cuda_version: str | None = None
    amd_available: bool = False
    rocm_available: bool = False
    rocm_version: str | None = None
    intel_available: bool = False
    gpus: list[GPUInfo] = Field(default_factory=list)



"""Tests for container backup management."""

import tempfile
import threading
import time
from datetime import datetime, timedelta
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest

from gaol.backup.exceptions import (
    BackupError,
    BackupExecutionError,
    BackupNotFoundError,
    ContainerNotFoundError,
    IncrementalNotAvailableError,
    PolicyNotFoundError,
    RestoreError,
    RetentionError,
    SchedulerError,
    StorageError,
    StorageNotConfiguredError,
    VerificationError,
)
from gaol.backup.models import (
    BackupHistory,
    BackupMetadata,
    BackupPolicy,
    BackupResult,
    BackupSchedule,
    BackupScheduleType,
    BackupStatus,
    BackupType,
    GlobalBackupConfig,
    RestoreResult,
    RetentionPolicy,
    SchedulerState,
    StorageBackendType,
    StorageConfig,
)
from gaol.backup.retention import RetentionEnforcer
from gaol.backup.scheduler import BackupScheduler
from gaol.backup.storage import LocalStorageBackend, get_storage_backend
from gaol.backup.store import BackupStore


class TestBackupModels:
    """Tests for backup Pydantic models."""

    def test_backup_schedule_defaults(self) -> None:
        """Test BackupSchedule default values."""
        schedule = BackupSchedule()
        assert schedule.schedule_type == BackupScheduleType.MANUAL
        assert schedule.time_of_day == "02:00"
        assert schedule.day_of_week == 0
        assert schedule.day_of_month == 1
        assert schedule.enabled is True

    def test_backup_schedule_custom(self) -> None:
        """Test BackupSchedule with custom values."""
        schedule = BackupSchedule(
            schedule_type=BackupScheduleType.WEEKLY,
            time_of_day="04:30",
            day_of_week=5,
            enabled=False,
        )
        assert schedule.schedule_type == BackupScheduleType.WEEKLY
        assert schedule.time_of_day == "04:30"
        assert schedule.day_of_week == 5
        assert schedule.enabled is False

    def test_backup_schedule_time_validation(self) -> None:
        """Test time format validation."""
        # Valid times
        BackupSchedule(time_of_day="00:00")
        BackupSchedule(time_of_day="23:59")
        BackupSchedule(time_of_day="12:30")

        # Invalid times
        with pytest.raises(ValueError):
            BackupSchedule(time_of_day="25:00")
        with pytest.raises(ValueError):
            BackupSchedule(time_of_day="12:60")

    def test_retention_policy_defaults(self) -> None:
        """Test RetentionPolicy default values."""
        policy = RetentionPolicy()
        assert policy.keep_last == 7
        assert policy.keep_daily == 7
        assert policy.keep_weekly == 4
        assert policy.keep_monthly == 6
        assert policy.max_age_days == 90
        assert policy.max_total_size_mb is None

    def test_retention_policy_custom(self) -> None:
        """Test RetentionPolicy with custom values."""
        policy = RetentionPolicy(
            keep_last=10,
            keep_daily=14,
            keep_weekly=8,
            keep_monthly=12,
            max_age_days=180,
            max_total_size_mb=10240,
        )
        assert policy.keep_last == 10
        assert policy.keep_daily == 14
        assert policy.max_total_size_mb == 10240

    def test_storage_config_local(self) -> None:
        """Test StorageConfig for local storage."""
        config = StorageConfig(
            backend=StorageBackendType.LOCAL,
            path="/backups",
        )
        assert config.backend == StorageBackendType.LOCAL
        assert config.path == "/backups"
        assert config.bucket is None

    def test_storage_config_s3(self) -> None:
        """Test StorageConfig for S3 storage."""
        config = StorageConfig(
            backend=StorageBackendType.S3,
            bucket="my-backups",
            region="us-east-1",
            prefix="containers/",
        )
        assert config.backend == StorageBackendType.S3
        assert config.bucket == "my-backups"
        assert config.region == "us-east-1"
        assert config.prefix == "containers/"

    def test_backup_policy_defaults(self) -> None:
        """Test BackupPolicy default values."""
        policy = BackupPolicy(container_name="test")
        assert policy.container_name == "test"
        assert policy.schedule.schedule_type == BackupScheduleType.MANUAL
        assert policy.retention.keep_last == 7
        assert policy.backup_type == BackupType.FULL
        assert policy.prefer_incremental is True
        assert policy.compression == "gzip"
        assert policy.verify_after_backup is True
        assert policy.notify_on_failure is True

    def test_backup_policy_custom(self) -> None:
        """Test BackupPolicy with custom values."""
        schedule = BackupSchedule(
            schedule_type=BackupScheduleType.DAILY,
            time_of_day="03:00",
        )
        retention = RetentionPolicy(keep_last=14)
        policy = BackupPolicy(
            container_name="mycontainer",
            schedule=schedule,
            retention=retention,
            backup_type=BackupType.INCREMENTAL,
            compression="xz",
            pre_backup_commands=["sync"],
            post_backup_commands=["echo done"],
        )
        assert policy.container_name == "mycontainer"
        assert policy.schedule.schedule_type == BackupScheduleType.DAILY
        assert policy.retention.keep_last == 14
        assert policy.backup_type == BackupType.INCREMENTAL
        assert policy.compression == "xz"
        assert policy.pre_backup_commands == ["sync"]

    def test_backup_metadata(self) -> None:
        """Test BackupMetadata model."""
        now = datetime.now()
        meta = BackupMetadata(
            backup_id="abc-123",
            container_name="test",
            backup_name="daily-backup",
            backup_type=BackupType.FULL,
            created_at=now,
            size_bytes=1024 * 1024,
            storage_backend=StorageBackendType.LOCAL,
            storage_path="/backups/test/daily-backup.tar.gz",
            checksum="sha256:abc123",
            verified=False,
        )
        assert meta.backup_id == "abc-123"
        assert meta.backup_name == "daily-backup"
        assert meta.backup_type == BackupType.FULL
        assert meta.size_bytes == 1024 * 1024
        assert meta.verified is False

    def test_backup_result_defaults(self) -> None:
        """Test BackupResult default values."""
        result = BackupResult(
            container_name="test",
            backup_name="test-backup",
            backup_id="abc-123",
            backup_type=BackupType.FULL,
        )
        assert result.container_name == "test"
        assert result.status == BackupStatus.PENDING
        assert result.size_bytes == 0
        assert result.error_message is None

    def test_backup_result_success(self) -> None:
        """Test BackupResult success property."""
        result = BackupResult(
            container_name="test",
            backup_name="test-backup",
            backup_id="abc-123",
            backup_type=BackupType.FULL,
            status=BackupStatus.COMPLETED,
        )
        assert result.success is True

        result2 = BackupResult(
            container_name="test",
            backup_name="test-backup",
            backup_id="abc-123",
            backup_type=BackupType.FULL,
            status=BackupStatus.FAILED,
        )
        assert result2.success is False

    def test_backup_result_duration(self) -> None:
        """Test BackupResult duration calculation."""
        start = datetime.now()
        end = start + timedelta(seconds=60)
        result = BackupResult(
            container_name="test",
            backup_name="test-backup",
            backup_id="abc-123",
            backup_type=BackupType.FULL,
            started_at=start,
            completed_at=end,
        )
        assert result.duration_seconds == pytest.approx(60.0, abs=0.1)

    def test_restore_result(self) -> None:
        """Test RestoreResult model."""
        result = RestoreResult(
            container_name="test",
            backup_name="daily-backup",
            backup_id="abc-123",
            status=BackupStatus.COMPLETED,
            pre_restore_snapshot="pre-restore-20240113",
        )
        assert result.container_name == "test"
        assert result.backup_name == "daily-backup"
        assert result.status == BackupStatus.COMPLETED
        assert result.pre_restore_snapshot == "pre-restore-20240113"

    def test_backup_history(self) -> None:
        """Test BackupHistory model."""
        history = BackupHistory(container_name="test")
        assert history.total_backups == 0
        assert history.successful_backups == 0
        assert history.failed_backups == 0

    def test_backup_history_add_backup(self) -> None:
        """Test BackupHistory adding backups."""
        history = BackupHistory(container_name="test")

        # Successful backup has completed_at set
        meta1 = BackupMetadata(
            backup_id="abc-123",
            container_name="test",
            backup_name="backup-1",
            backup_type=BackupType.FULL,
            created_at=datetime.now(),
            completed_at=datetime.now(),  # This makes it successful
            size_bytes=1024,
            storage_backend=StorageBackendType.LOCAL,
            storage_path="/test",
        )
        history.add_backup(meta1)
        assert history.total_backups == 1
        assert history.successful_backups == 1
        assert history.last_backup is not None

        # Failed backup has no completed_at
        meta2 = BackupMetadata(
            backup_id="def-456",
            container_name="test",
            backup_name="backup-2",
            backup_type=BackupType.FULL,
            created_at=datetime.now(),
            size_bytes=2048,
            storage_backend=StorageBackendType.LOCAL,
            storage_path="/test2",
        )
        history.add_backup(meta2)
        assert history.total_backups == 2
        assert history.failed_backups == 1

    def test_backup_history_remove_backup(self) -> None:
        """Test BackupHistory removing backups."""
        history = BackupHistory(container_name="test")

        meta = BackupMetadata(
            backup_id="abc-123",
            container_name="test",
            backup_name="backup-1",
            backup_type=BackupType.FULL,
            created_at=datetime.now(),
            completed_at=datetime.now(),
            size_bytes=1024,
            storage_backend=StorageBackendType.LOCAL,
            storage_path="/test",
        )
        history.add_backup(meta)
        assert history.total_backups == 1
        assert len(history.backups) == 1

        # remove_backup removes from list but doesn't decrement counters
        result = history.remove_backup("abc-123")
        assert result is True
        assert len(history.backups) == 0

    def test_global_backup_config(self) -> None:
        """Test GlobalBackupConfig model."""
        config = GlobalBackupConfig()
        assert config.default_storage.backend == StorageBackendType.LOCAL
        assert config.scheduler_enabled is True
        assert config.parallel_backups == 1

    def test_scheduler_state(self) -> None:
        """Test SchedulerState model."""
        state = SchedulerState()
        assert state.running is False
        assert state.started_at is None
        assert state.next_scheduled == {}


class TestBackupExceptions:
    """Tests for backup exceptions."""

    def test_backup_error_base(self) -> None:
        """Test base BackupError."""
        err = BackupError("test error")
        assert err.message == "test error"
        assert str(err) == "test error"

    def test_container_not_found_error(self) -> None:
        """Test ContainerNotFoundError."""
        err = ContainerNotFoundError("mycontainer")
        assert err.container_name == "mycontainer"
        assert "mycontainer" in str(err)

    def test_backup_not_found_error(self) -> None:
        """Test BackupNotFoundError."""
        err = BackupNotFoundError("mycontainer", "daily-backup")
        assert err.container_name == "mycontainer"
        assert err.backup_name == "daily-backup"
        assert "daily-backup" in str(err)

    def test_policy_not_found_error(self) -> None:
        """Test PolicyNotFoundError."""
        err = PolicyNotFoundError("mycontainer")
        assert err.container_name == "mycontainer"
        assert "mycontainer" in str(err)

    def test_storage_error(self) -> None:
        """Test StorageError."""
        err = StorageError("local", "upload", "disk full")
        assert err.backend == "local"
        assert err.operation == "upload"
        assert err.reason == "disk full"
        assert "upload" in str(err)

    def test_storage_not_configured_error(self) -> None:
        """Test StorageNotConfiguredError."""
        err = StorageNotConfiguredError("S3")
        assert err.backend == "S3"
        assert "S3" in str(err)

    def test_backup_execution_error(self) -> None:
        """Test BackupExecutionError."""
        err = BackupExecutionError("export failed", output="error log")
        assert err.message == "export failed"
        assert err.output == "error log"

    def test_restore_error(self) -> None:
        """Test RestoreError."""
        err = RestoreError("mycontainer", "daily-backup", "checksum mismatch")
        assert err.container_name == "mycontainer"
        assert err.backup_name == "daily-backup"
        assert err.reason == "checksum mismatch"

    def test_retention_error(self) -> None:
        """Test RetentionError."""
        err = RetentionError("cannot delete last backup")
        assert err.message == "cannot delete last backup"
        assert "cannot delete last backup" in str(err)

    def test_verification_error(self) -> None:
        """Test VerificationError."""
        err = VerificationError("daily-backup", "checksum mismatch")
        assert err.backup_name == "daily-backup"
        assert err.reason == "checksum mismatch"

    def test_incremental_not_available_error(self) -> None:
        """Test IncrementalNotAvailableError."""
        err = IncrementalNotAvailableError("mycontainer", "no CoW support")
        assert err.container_name == "mycontainer"
        assert err.reason == "no CoW support"

    def test_scheduler_error(self) -> None:
        """Test SchedulerError."""
        err = SchedulerError("scheduler already running")
        assert err.message == "scheduler already running"
        assert "scheduler already running" in str(err)


class TestBackupStore:
    """Tests for BackupStore."""

    def test_store_init(self) -> None:
        """Test BackupStore initialization."""
        with tempfile.TemporaryDirectory() as tmpdir:
            store = BackupStore(Path(tmpdir))
            assert store._store_dir == Path(tmpdir)

    def test_store_save_load_policy(self) -> None:
        """Test saving and loading backup policy."""
        with tempfile.TemporaryDirectory() as tmpdir:
            store = BackupStore(Path(tmpdir))

            policy = BackupPolicy(
                container_name="test",
                backup_type=BackupType.FULL,
            )
            store.save_policy(policy)

            loaded = store.get_policy("test")
            assert loaded is not None
            assert loaded.container_name == "test"
            assert loaded.backup_type == BackupType.FULL

    def test_store_delete_policy(self) -> None:
        """Test deleting backup policy."""
        with tempfile.TemporaryDirectory() as tmpdir:
            store = BackupStore(Path(tmpdir))

            policy = BackupPolicy(container_name="test")
            store.save_policy(policy)

            assert store.delete_policy("test") is True
            assert store.get_policy("test") is None
            assert store.delete_policy("test") is False

    def test_store_list_policies(self) -> None:
        """Test listing backup policies."""
        with tempfile.TemporaryDirectory() as tmpdir:
            store = BackupStore(Path(tmpdir))

            store.save_policy(BackupPolicy(container_name="test1"))
            store.save_policy(BackupPolicy(container_name="test2"))

            policies = store.list_policies()
            assert len(policies) == 2
            names = {p.container_name for p in policies}
            assert names == {"test1", "test2"}

    def test_store_save_load_history(self) -> None:
        """Test saving and loading backup history."""
        with tempfile.TemporaryDirectory() as tmpdir:
            store = BackupStore(Path(tmpdir))

            history = BackupHistory(container_name="test")
            meta = BackupMetadata(
                backup_id="abc-123",
                container_name="test",
                backup_name="backup-1",
                backup_type=BackupType.FULL,
                created_at=datetime.now(),
                size_bytes=1024,
                storage_backend=StorageBackendType.LOCAL,
                storage_path="/test",
            )
            history.add_backup(meta)
            store.save_history(history)

            loaded = store.get_history("test")
            assert loaded.container_name == "test"
            assert loaded.total_backups == 1

    def test_store_save_load_metadata(self) -> None:
        """Test saving and loading backup metadata."""
        with tempfile.TemporaryDirectory() as tmpdir:
            store = BackupStore(Path(tmpdir))

            meta = BackupMetadata(
                backup_id="abc-123",
                container_name="test",
                backup_name="daily-backup",
                backup_type=BackupType.FULL,
                created_at=datetime.now(),
                size_bytes=1024,
                storage_backend=StorageBackendType.LOCAL,
                storage_path="/test",
            )
            store.save_backup_metadata(meta)

            loaded = store.get_backup_metadata("test", "abc-123")
            assert loaded is not None
            assert loaded.backup_id == "abc-123"
            assert loaded.backup_name == "daily-backup"

    def test_store_get_backup_by_name(self) -> None:
        """Test getting backup metadata by name."""
        with tempfile.TemporaryDirectory() as tmpdir:
            store = BackupStore(Path(tmpdir))

            meta = BackupMetadata(
                backup_id="abc-123",
                container_name="test",
                backup_name="daily-backup",
                backup_type=BackupType.FULL,
                created_at=datetime.now(),
                size_bytes=1024,
                storage_backend=StorageBackendType.LOCAL,
                storage_path="/test",
            )
            store.save_backup_metadata(meta)

            loaded = store.get_backup_by_name("test", "daily-backup")
            assert loaded is not None
            assert loaded.backup_id == "abc-123"

            # Test non-existent backup
            assert store.get_backup_by_name("test", "nonexistent") is None

    def test_store_list_backups(self) -> None:
        """Test listing backups for a container."""
        with tempfile.TemporaryDirectory() as tmpdir:
            store = BackupStore(Path(tmpdir))

            for i in range(3):
                meta = BackupMetadata(
                    backup_id=f"id-{i}",
                    container_name="test",
                    backup_name=f"backup-{i}",
                    backup_type=BackupType.FULL,
                    created_at=datetime.now() + timedelta(hours=i),
                    size_bytes=1024,
                    storage_backend=StorageBackendType.LOCAL,
                    storage_path=f"/test/{i}",
                )
                store.save_backup_metadata(meta)

            backups = store.list_backups("test")
            assert len(backups) == 3
            # Should be sorted by creation date (newest first)
            assert backups[0].backup_name == "backup-2"
            assert backups[2].backup_name == "backup-0"

    def test_store_delete_backup_metadata(self) -> None:
        """Test deleting backup metadata."""
        with tempfile.TemporaryDirectory() as tmpdir:
            store = BackupStore(Path(tmpdir))

            meta = BackupMetadata(
                backup_id="abc-123",
                container_name="test",
                backup_name="daily-backup",
                backup_type=BackupType.FULL,
                created_at=datetime.now(),
                size_bytes=1024,
                storage_backend=StorageBackendType.LOCAL,
                storage_path="/test",
            )
            store.save_backup_metadata(meta)

            assert store.delete_backup_metadata("test", "abc-123") is True
            assert store.get_backup_metadata("test", "abc-123") is None
            assert store.delete_backup_metadata("test", "abc-123") is False

    def test_store_global_config(self) -> None:
        """Test global backup configuration."""
        with tempfile.TemporaryDirectory() as tmpdir:
            store = BackupStore(Path(tmpdir))

            # Get default config
            config = store.get_global_config()
            assert config.default_storage.backend == StorageBackendType.LOCAL

            # Save custom config
            config.default_storage = StorageConfig(
                backend=StorageBackendType.S3,
                bucket="my-bucket",
            )
            store.save_global_config(config)

            # Reload and verify
            loaded = store.get_global_config()
            assert loaded.default_storage.backend == StorageBackendType.S3
            assert loaded.default_storage.bucket == "my-bucket"


class TestLocalStorageBackend:
    """Tests for LocalStorageBackend."""

    def test_local_storage_upload_download(self) -> None:
        """Test uploading and downloading files."""
        with tempfile.TemporaryDirectory() as tmpdir:
            storage = LocalStorageBackend(Path(tmpdir))

            # Create test file
            test_content = b"test backup content"
            source = Path(tmpdir) / "source.tar.gz"
            source.write_bytes(test_content)

            # Upload
            remote_path = storage.upload(source, "container/backup.tar.gz")
            assert storage.exists(remote_path)

            # Download
            dest = Path(tmpdir) / "downloaded.tar.gz"
            storage.download(remote_path, dest)
            assert dest.read_bytes() == test_content

    def test_local_storage_delete(self) -> None:
        """Test deleting files."""
        with tempfile.TemporaryDirectory() as tmpdir:
            storage = LocalStorageBackend(Path(tmpdir))

            # Create and upload test file
            source = Path(tmpdir) / "source.tar.gz"
            source.write_bytes(b"test content")
            remote_path = storage.upload(source, "container/backup.tar.gz")

            assert storage.exists(remote_path)
            storage.delete(remote_path)
            assert not storage.exists(remote_path)

    def test_local_storage_list(self) -> None:
        """Test listing files."""
        with tempfile.TemporaryDirectory() as tmpdir:
            storage = LocalStorageBackend(Path(tmpdir))

            # Create test files
            for i in range(3):
                source = Path(tmpdir) / f"source{i}.tar.gz"
                source.write_bytes(f"content {i}".encode())
                storage.upload(source, f"container/backup{i}.tar.gz")

            files = storage.list("container/")
            assert len(files) == 3

    def test_local_storage_get_size(self) -> None:
        """Test getting file size."""
        with tempfile.TemporaryDirectory() as tmpdir:
            storage = LocalStorageBackend(Path(tmpdir))

            content = b"test content 1234567890"
            source = Path(tmpdir) / "source.tar.gz"
            source.write_bytes(content)
            remote_path = storage.upload(source, "container/backup.tar.gz")

            size = storage.get_size(remote_path)
            assert size == len(content)


class TestRetentionEnforcer:
    """Tests for RetentionEnforcer."""

    def test_retention_keep_last(self) -> None:
        """Test keep_last retention rule."""
        with tempfile.TemporaryDirectory() as tmpdir:
            store = BackupStore(Path(tmpdir) / "store")
            storage = LocalStorageBackend(Path(tmpdir) / "storage")
            enforcer = RetentionEnforcer(store, storage)

            # Create 10 backups
            for i in range(10):
                meta = BackupMetadata(
                    backup_id=f"id-{i}",
                    container_name="test",
                    backup_name=f"backup-{i}",
                    backup_type=BackupType.FULL,
                    created_at=datetime.now() - timedelta(days=9 - i),
                    size_bytes=1024,
                    storage_backend=StorageBackendType.LOCAL,
                    storage_path=f"test/backup-{i}.tar.gz",
                )
                store.save_backup_metadata(meta)

            # Apply retention: keep last 5
            policy = RetentionPolicy(
                keep_last=5,
                keep_daily=None,
                keep_weekly=None,
                keep_monthly=None,
                max_age_days=None,
            )

            deleted = enforcer.apply_retention("test", policy, dry_run=True)
            assert len(deleted) == 5  # Should delete 5 oldest

    def test_retention_max_age(self) -> None:
        """Test max_age_days retention rule."""
        with tempfile.TemporaryDirectory() as tmpdir:
            store = BackupStore(Path(tmpdir) / "store")
            storage = LocalStorageBackend(Path(tmpdir) / "storage")
            enforcer = RetentionEnforcer(store, storage)

            # Create backups with different ages
            for i in range(5):
                meta = BackupMetadata(
                    backup_id=f"id-{i}",
                    container_name="test",
                    backup_name=f"backup-{i}",
                    backup_type=BackupType.FULL,
                    created_at=datetime.now() - timedelta(days=i * 30),
                    size_bytes=1024,
                    storage_backend=StorageBackendType.LOCAL,
                    storage_path=f"test/backup-{i}.tar.gz",
                )
                store.save_backup_metadata(meta)

            # Apply retention: max 60 days
            policy = RetentionPolicy(
                keep_last=None,
                keep_daily=None,
                keep_weekly=None,
                keep_monthly=None,
                max_age_days=60,
            )

            deleted = enforcer.apply_retention("test", policy, dry_run=True)
            # Backups older than 60 days should be deleted (days 90, 120)
            assert len(deleted) >= 2

    def test_retention_preview(self) -> None:
        """Test retention policy preview."""
        with tempfile.TemporaryDirectory() as tmpdir:
            store = BackupStore(Path(tmpdir) / "store")
            storage = LocalStorageBackend(Path(tmpdir) / "storage")
            enforcer = RetentionEnforcer(store, storage)

            # Create backups
            for i in range(5):
                meta = BackupMetadata(
                    backup_id=f"id-{i}",
                    container_name="test",
                    backup_name=f"backup-{i}",
                    backup_type=BackupType.FULL,
                    created_at=datetime.now() - timedelta(days=i),
                    size_bytes=1024,
                    storage_backend=StorageBackendType.LOCAL,
                    storage_path=f"test/backup-{i}.tar.gz",
                )
                store.save_backup_metadata(meta)

            # Disable all other retention rules, only keep_last=3
            policy = RetentionPolicy(
                keep_last=3,
                keep_daily=None,
                keep_weekly=None,
                keep_monthly=None,
                max_age_days=None,
            )
            to_keep, to_delete = enforcer.get_retention_preview("test", policy)

            assert len(to_keep) == 3
            assert len(to_delete) == 2


class TestBackupScheduler:
    """Tests for BackupScheduler."""

    def test_scheduler_start_stop(self) -> None:
        """Test starting and stopping scheduler."""
        with tempfile.TemporaryDirectory() as tmpdir:
            store = BackupStore(Path(tmpdir))
            scheduler = BackupScheduler(store)

            assert not scheduler.is_running

            scheduler.start()
            assert scheduler.is_running
            assert scheduler.state.running is True

            scheduler.stop()
            assert not scheduler.is_running
            assert scheduler.state.running is False

    def test_scheduler_calculate_next_time_daily(self) -> None:
        """Test calculating next daily backup time."""
        with tempfile.TemporaryDirectory() as tmpdir:
            store = BackupStore(Path(tmpdir))
            scheduler = BackupScheduler(store)

            schedule = BackupSchedule(
                schedule_type=BackupScheduleType.DAILY,
                time_of_day="03:00",
            )

            next_time = scheduler._calculate_next_scheduled_time(schedule)
            assert next_time is not None
            assert next_time.hour == 3
            assert next_time.minute == 0

    def test_scheduler_calculate_next_time_weekly(self) -> None:
        """Test calculating next weekly backup time."""
        with tempfile.TemporaryDirectory() as tmpdir:
            store = BackupStore(Path(tmpdir))
            scheduler = BackupScheduler(store)

            schedule = BackupSchedule(
                schedule_type=BackupScheduleType.WEEKLY,
                time_of_day="02:00",
                day_of_week=6,  # Sunday
            )

            next_time = scheduler._calculate_next_scheduled_time(schedule)
            assert next_time is not None
            assert next_time.weekday() == 6
            assert next_time.hour == 2

    def test_scheduler_trigger_backup(self) -> None:
        """Test manually triggering a backup."""
        with tempfile.TemporaryDirectory() as tmpdir:
            store = BackupStore(Path(tmpdir))
            callback_called = []

            def mock_backup(container_name: str) -> BackupResult:
                callback_called.append(container_name)
                return BackupResult(
                    container_name=container_name,
                    backup_name="triggered-backup",
                    backup_id="test-id",
                    backup_type=BackupType.FULL,
                    status=BackupStatus.COMPLETED,
                )

            scheduler = BackupScheduler(store, on_backup=mock_backup)

            result = scheduler.trigger_backup("test")
            assert result is not None
            assert result.status == BackupStatus.COMPLETED
            assert "test" in callback_called

    def test_scheduler_remove_container(self) -> None:
        """Test removing container from scheduler state."""
        with tempfile.TemporaryDirectory() as tmpdir:
            store = BackupStore(Path(tmpdir))
            scheduler = BackupScheduler(store)

            # Manually add to state
            scheduler._state.next_scheduled["test"] = datetime.now()
            assert "test" in scheduler.state.next_scheduled

            scheduler.remove_container("test")
            assert "test" not in scheduler.state.next_scheduled


class TestStorageBackendFactory:
    """Tests for storage backend factory."""

    def test_get_storage_backend_default(self) -> None:
        """Test getting default storage backend."""
        backend = get_storage_backend(None)
        assert isinstance(backend, LocalStorageBackend)

    def test_get_storage_backend_local(self) -> None:
        """Test getting local storage backend."""
        config = StorageConfig(
            backend=StorageBackendType.LOCAL,
            path="/custom/path",
        )
        backend = get_storage_backend(config)
        assert isinstance(backend, LocalStorageBackend)

    def test_get_storage_backend_s3(self) -> None:
        """Test getting S3 storage backend configuration."""
        config = StorageConfig(
            backend=StorageBackendType.S3,
            bucket="test-bucket",
            region="us-east-1",
        )
        # This will create an S3 backend (may not connect without credentials)
        backend = get_storage_backend(config)
        assert backend.name == "s3"


class TestBackupTypes:
    """Tests for backup type enumerations."""

    def test_backup_schedule_type_values(self) -> None:
        """Test BackupScheduleType values."""
        assert BackupScheduleType.MANUAL.value == "manual"
        assert BackupScheduleType.DAILY.value == "daily"
        assert BackupScheduleType.WEEKLY.value == "weekly"
        assert BackupScheduleType.MONTHLY.value == "monthly"

    def test_backup_status_values(self) -> None:
        """Test BackupStatus values."""
        assert BackupStatus.PENDING.value == "pending"
        assert BackupStatus.PREPARING.value == "preparing"
        assert BackupStatus.SNAPSHOTTING.value == "snapshotting"
        assert BackupStatus.UPLOADING.value == "uploading"
        assert BackupStatus.VERIFYING.value == "verifying"
        assert BackupStatus.COMPLETED.value == "completed"
        assert BackupStatus.FAILED.value == "failed"

    def test_backup_type_values(self) -> None:
        """Test BackupType values."""
        assert BackupType.FULL.value == "full"
        assert BackupType.INCREMENTAL.value == "incremental"
        assert BackupType.CONFIG_ONLY.value == "config_only"

    def test_storage_backend_type_values(self) -> None:
        """Test StorageBackendType values."""
        assert StorageBackendType.LOCAL.value == "local"
        assert StorageBackendType.S3.value == "s3"



"""Scheduled backup functionality for gaol containers.

This module provides automatic backup functionality including:
- Full and incremental backups
- Automatic backup scheduling (daily, weekly, monthly)
- Retention policies (keep last N, keep daily/weekly/monthly)
- Cloud storage support (local, S3-compatible)
- Backup verification and restore

Example:
    >>> from gaol.backup import get_backup_manager
    >>> manager = get_backup_manager()
    >>> result = manager.create("mycontainer", name="daily-backup")
    >>> manager.set_schedule("mycontainer", BackupScheduleType.DAILY, "02:00")
    >>> manager.restore("mycontainer", "daily-backup")
"""

from gaol.backup.exceptions import (
    BackupError,
    BackupExecutionError,
    BackupNotFoundError,
    ContainerNotFoundError,
    IncrementalNotAvailableError,
    PolicyNotFoundError,
    RestoreError,
    RetentionError,
    SchedulerError,
    StorageError,
    StorageNotConfiguredError,
    VerificationError,
)
from gaol.backup.manager import BackupManager, get_backup_manager
from gaol.backup.models import (
    BackupHistory,
    BackupMetadata,
    BackupPolicy,
    BackupResult,
    BackupSchedule,
    BackupScheduleType,
    BackupStatus,
    BackupType,
    GlobalBackupConfig,
    RestoreResult,
    RetentionPolicy,
    SchedulerState,
    StorageBackendType,
    StorageConfig,
)
from gaol.backup.retention import RetentionEnforcer
from gaol.backup.scheduler import BackupScheduler, get_backup_scheduler
from gaol.backup.storage import (
    LocalStorageBackend,
    S3StorageBackend,
    StorageBackend,
    get_storage_backend,
)
from gaol.backup.store import BackupStore, get_backup_store

__all__ = [
    # Manager
    "BackupManager",
    "get_backup_manager",
    # Models
    "BackupScheduleType",
    "BackupStatus",
    "BackupType",
    "StorageBackendType",
    "BackupSchedule",
    "RetentionPolicy",
    "StorageConfig",
    "BackupPolicy",
    "BackupMetadata",
    "BackupResult",
    "RestoreResult",
    "BackupHistory",
    "GlobalBackupConfig",
    "SchedulerState",
    # Store
    "BackupStore",
    "get_backup_store",
    # Storage
    "StorageBackend",
    "LocalStorageBackend",
    "S3StorageBackend",
    "get_storage_backend",
    # Scheduler
    "BackupScheduler",
    "get_backup_scheduler",
    # Retention
    "RetentionEnforcer",
    # Exceptions
    "BackupError",
    "ContainerNotFoundError",
    "BackupNotFoundError",
    "PolicyNotFoundError",
    "StorageError",
    "StorageNotConfiguredError",
    "BackupExecutionError",
    "RestoreError",
    "RetentionError",
    "VerificationError",
    "IncrementalNotAvailableError",
    "SchedulerError",
]



"""Container context management for shell integration.

This module provides utilities for managing the current container context
in shell environments, enabling prompt integration and quick access commands.
"""

from __future__ import annotations

import os

from gaol.models import ContainerStatus
from gaol.runtimes import get_runtime
from gaol.store import get_container_store


class ContainerContext:
    """Manage container context for shell integration.

    The container context is tracked via the GAOL_CONTAINER
    environment variable, which can be set by shell helper functions.

    Example:
        >>> context = ContainerContext()
        >>> context.get_current()
        'mycontainer'
        >>> context.get_status_short('mycontainer')
        'running'
    """

    ENV_VAR = "GAOL_CONTAINER"

    @classmethod
    def get_current(cls) -> str | None:
        """Get current container from environment.

        Returns:
            Container name if set, None otherwise.
        """
        return os.environ.get(cls.ENV_VAR)

    @classmethod
    def get_status_short(cls, container_name: str) -> str:
        """Get short status string for prompt display.

        Args:
            container_name: Name of the container.

        Returns:
            Short status string (e.g., "running", "stopped", "?").
        """
        store = get_container_store()
        stored = store.get(container_name)

        if not stored:
            return "?"

        try:
            rt = get_runtime(stored.runtime)
            state = rt.status(container_name)
            if state:
                return state.status.value
        except Exception:
            pass

        return "?"

    @classmethod
    def get_context_info(cls, container_name: str) -> dict[str, str | None]:
        """Get detailed context information for a container.

        Args:
            container_name: Name of the container.

        Returns:
            Dictionary with container info (name, runtime, status, image, ip).
        """
        store = get_container_store()
        stored = store.get(container_name)

        info: dict[str, str | None] = {
            "name": container_name,
            "runtime": None,
            "status": None,
            "image": None,
            "ip_address": None,
        }

        if not stored:
            return info

        info["runtime"] = stored.runtime
        info["image"] = stored.config.image if stored.config else None

        try:
            rt = get_runtime(stored.runtime)
            state = rt.status(container_name)
            if state:
                info["status"] = state.status.value
                info["ip_address"] = state.ip_address
        except Exception:
            pass

        return info

    @classmethod
    def list_containers(cls) -> list[str]:
        """List all known container names.

        Returns:
            List of container names.
        """
        store = get_container_store()
        return store.list()

    @classmethod
    def container_exists(cls, container_name: str) -> bool:
        """Check if a container exists in the store.

        Args:
            container_name: Name of the container.

        Returns:
            True if container exists.
        """
        store = get_container_store()
        return store.get(container_name) is not None



"""Networking module for gaol container manager.

Provides port conflict detection, DNS configuration, bridge management,
and firewall rules for port forwarding.
"""

from gaol.networking.bridge import (
    BridgeError,
    BridgeInfo,
    bridge_exists,
    create_bridge,
    delete_bridge,
    ensure_bridge,
    get_bridge_info,
    list_bridges,
)
from gaol.networking.dns import (
    PUBLIC_DNS_SERVERS,
    DNSError,
    InvalidDNSServerError,
    generate_resolv_conf,
    get_default_dns_config,
    get_system_dns,
    parse_resolv_conf,
    validate_dns_server,
)
from gaol.networking.firewall import (
    FirewallBackend,
    FirewallError,
    FirewallManager,
    ForwardingRule,
    add_port_forward,
    remove_port_forward,
)
from gaol.networking.manager import (
    NetworkManager,
    get_network_manager,
)
from gaol.networking.models import (
    BridgeConfig,
    DNSConfig,
    IPVersion,
    NetworkConfigV2,
    PortMapping,
    PortMappingList,
    Protocol,
)
from gaol.networking.ports import (
    PortConflictError,
    PortError,
    PortInUseError,
    check_port_conflicts,
    find_available_port,
    get_port_process,
    is_port_available,
    validate_port_mappings,
)

__all__ = [
    # Models
    "BridgeConfig",
    "DNSConfig",
    "IPVersion",
    "NetworkConfigV2",
    "PortMapping",
    "PortMappingList",
    "Protocol",
    # Ports
    "PortConflictError",
    "PortError",
    "PortInUseError",
    "check_port_conflicts",
    "find_available_port",
    "get_port_process",
    "is_port_available",
    "validate_port_mappings",
    # DNS
    "DNSError",
    "InvalidDNSServerError",
    "PUBLIC_DNS_SERVERS",
    "generate_resolv_conf",
    "get_default_dns_config",
    "get_system_dns",
    "parse_resolv_conf",
    "validate_dns_server",
    # Bridge
    "BridgeError",
    "BridgeInfo",
    "bridge_exists",
    "create_bridge",
    "delete_bridge",
    "ensure_bridge",
    "get_bridge_info",
    "list_bridges",
    # Firewall
    "FirewallBackend",
    "FirewallError",
    "FirewallManager",
    "ForwardingRule",
    "add_port_forward",
    "remove_port_forward",
    # Manager
    "NetworkManager",
    "get_network_manager",
]



"""Global configuration management for gaol."""

from __future__ import annotations

import os
from pathlib import Path
from typing import Annotated

import yaml
from pydantic import BaseModel, Field


def _get_user_home() -> Path:
    """Get the real user's home directory, even under sudo."""
    sudo_user = os.environ.get("SUDO_USER")
    if sudo_user:
        import pwd

        try:
            return Path(pwd.getpwnam(sudo_user).pw_dir)
        except KeyError:
            pass
    return Path.home()


def get_config_dir() -> Path:
    """Get the configuration directory."""
    xdg_config = os.environ.get("XDG_CONFIG_HOME")
    if xdg_config:
        config_dir = Path(xdg_config) / "gaol"
    else:
        config_dir = _get_user_home() / ".config" / "gaol"
    config_dir.mkdir(parents=True, exist_ok=True)
    return config_dir


def get_data_dir() -> Path:
    """Get the data directory."""
    xdg_data = os.environ.get("XDG_DATA_HOME")
    if xdg_data:
        data_dir = Path(xdg_data) / "gaol"
    else:
        data_dir = _get_user_home() / ".local" / "share" / "gaol"
    data_dir.mkdir(parents=True, exist_ok=True)
    return data_dir


class RuntimeConfig(BaseModel):
    """Configuration for a specific runtime."""

    enabled: bool = True
    default_image: str | None = None


class Config(BaseModel):
    """Global gaol configuration."""

    default_runtime: Annotated[str, Field(default="docker")]
    default_distro: Annotated[str, Field(default="trixie")]
    profile_dirs: Annotated[
        list[str],
        Field(
            default_factory=lambda: [str(get_config_dir() / "profiles"), "./profiles"],
            description="Directories to search for profile definitions",
        ),
    ]
    runtimes: Annotated[
        dict[str, RuntimeConfig],
        Field(
            default_factory=lambda: {
                "docker": RuntimeConfig(),
                "nspawn": RuntimeConfig(),
                "libvirt": RuntimeConfig(),
            }
        ),
    ]

    @classmethod
    def load(cls) -> Config:
        """Load configuration from file or return defaults."""
        config_file = get_config_dir() / "config.yaml"
        if config_file.exists():
            with open(config_file) as f:
                data = yaml.safe_load(f) or {}
            return cls(**data)
        return cls()

    def save(self) -> None:
        """Save configuration to file."""
        config_file = get_config_dir() / "config.yaml"
        with open(config_file, "w") as f:
            yaml.dump(self.model_dump(), f, default_flow_style=False)


# Global configuration instance
_config: Config | None = None


def get_config() -> Config:
    """Get the global configuration instance."""
    global _config
    if _config is None:
        _config = Config.load()
    return _config



"""Exceptions for template gallery operations."""

from __future__ import annotations


class TemplateError(Exception):
    """Base exception for template operations."""

    pass


class TemplateNotFoundError(TemplateError):
    """Template not found in any source."""

    def __init__(self, template_id: str, searched: list[str] | None = None) -> None:
        searched_str = ""
        if searched:
            searched_str = f" (searched: {', '.join(searched)})"
        super().__init__(f"Template not found: {template_id}{searched_str}")
        self.template_id = template_id
        self.searched = searched or []


class TemplateParseError(TemplateError):
    """Failed to parse template file."""

    def __init__(
        self,
        template_id: str,
        reason: str,
        cause: Exception | None = None,
    ) -> None:
        super().__init__(f"Failed to parse template '{template_id}': {reason}")
        self.template_id = template_id
        self.reason = reason
        self.cause = cause


class TemplateValidationError(TemplateError):
    """Template validation failed."""

    def __init__(
        self,
        template_id: str,
        errors: list[str],
    ) -> None:
        error_list = "\n  - ".join(errors)
        super().__init__(
            f"Template '{template_id}' validation failed:\n  - {error_list}"
        )
        self.template_id = template_id
        self.errors = errors


class TemplateExistsError(TemplateError):
    """Template already exists."""

    def __init__(self, template_id: str, location: str | None = None) -> None:
        msg = f"Template already exists: {template_id}"
        if location:
            msg += f" (at {location})"
        super().__init__(msg)
        self.template_id = template_id
        self.location = location


class GalleryError(TemplateError):
    """Base exception for gallery operations."""

    pass


class GalleryFetchError(GalleryError):
    """Failed to fetch from remote gallery."""

    def __init__(
        self,
        url: str,
        reason: str,
        cause: Exception | None = None,
    ) -> None:
        super().__init__(f"Failed to fetch from gallery: {reason}")
        self.url = url
        self.reason = reason
        self.cause = cause


class GalleryIndexError(GalleryError):
    """Failed to parse gallery index."""

    def __init__(
        self,
        reason: str,
        cause: Exception | None = None,
    ) -> None:
        super().__init__(f"Invalid gallery index: {reason}")
        self.reason = reason
        self.cause = cause


class GalleryOfflineError(GalleryError):
    """Gallery is not accessible."""

    def __init__(self, url: str, reason: str | None = None) -> None:
        msg = f"Gallery is not accessible: {url}"
        if reason:
            msg += f" ({reason})"
        super().__init__(msg)
        self.url = url
        self.reason = reason


class TemplateStoreError(TemplateError):
    """Error with template storage."""

    def __init__(self, operation: str, reason: str) -> None:
        super().__init__(f"Template store {operation} failed: {reason}")
        self.operation = operation
        self.reason = reason


class CategoryNotFoundError(TemplateError):
    """Category not found."""

    def __init__(self, category: str) -> None:
        super().__init__(f"Category not found: {category}")
        self.category = category



"""Tests for distro module."""

import pytest

from gaol.distro import (
    DistroFamily,
    DistroInfo,
    PackageManager,
    ServiceManager,
    get_cloud_image_url,
    get_debootstrap_cmd,
    get_distro_family,
    get_distro_info,
    get_docker_image,
    get_mirror,
    get_package_manager_commands,
    get_service_commands,
    list_supported_distros,
    translate_package,
    translate_packages,
    translate_service,
)


class TestDistroRegistry:
    """Tests for distro registry and info."""

    def test_list_supported_distros(self) -> None:
        """Test listing all supported distros."""
        distros = list_supported_distros()
        assert len(distros) >= 9
        assert "trixie" in distros
        assert "bookworm" in distros
        assert "noble" in distros
        assert "fedora" in distros
        assert "alpine" in distros
        assert "arch" in distros

    def test_get_distro_info_debian(self) -> None:
        """Test getting info for Debian distro."""
        info = get_distro_info("trixie")
        assert isinstance(info, DistroInfo)
        assert info.name == "trixie"
        assert info.family == DistroFamily.DEBIAN
        assert info.package_manager == PackageManager.APT
        assert info.service_manager == ServiceManager.SYSTEMD
        assert info.docker_image == "debian:trixie"
        assert info.cloud_image_url is not None
        assert info.debootstrap_cmd is not None

    def test_get_distro_info_ubuntu(self) -> None:
        """Test getting info for Ubuntu distro."""
        info = get_distro_info("noble")
        assert info.name == "noble"
        assert info.family == DistroFamily.DEBIAN
        assert info.package_manager == PackageManager.APT
        assert "ubuntu" in info.docker_image.lower()

    def test_get_distro_info_fedora(self) -> None:
        """Test getting info for Fedora distro."""
        info = get_distro_info("fedora")
        assert info.name == "fedora"
        assert info.family == DistroFamily.REDHAT
        assert info.package_manager == PackageManager.DNF
        assert info.service_manager == ServiceManager.SYSTEMD

    def test_get_distro_info_alpine(self) -> None:
        """Test getting info for Alpine distro."""
        info = get_distro_info("alpine")
        assert info.name == "alpine"
        assert info.family == DistroFamily.ALPINE
        assert info.package_manager == PackageManager.APK
        assert info.service_manager == ServiceManager.OPENRC

    def test_get_distro_info_arch(self) -> None:
        """Test getting info for Arch distro."""
        info = get_distro_info("arch")
        assert info.name == "arch"
        assert info.family == DistroFamily.ARCH
        assert info.package_manager == PackageManager.PACMAN
        assert info.service_manager == ServiceManager.SYSTEMD

    def test_get_distro_info_unknown(self) -> None:
        """Test getting info for unknown distro raises KeyError."""
        with pytest.raises(KeyError, match="Unknown distribution"):
            get_distro_info("unknown-distro")

    def test_get_distro_family(self) -> None:
        """Test getting distro family."""
        assert get_distro_family("trixie") == DistroFamily.DEBIAN
        assert get_distro_family("fedora") == DistroFamily.REDHAT
        assert get_distro_family("alpine") == DistroFamily.ALPINE
        assert get_distro_family("arch") == DistroFamily.ARCH


class TestPackageTranslation:
    """Tests for package name translation."""

    def test_translate_package_simple(self) -> None:
        """Test simple package translation."""
        # curl is the same across distros
        assert translate_package("curl", "trixie") == "curl"
        assert translate_package("curl", "fedora") == "curl"
        assert translate_package("curl", "alpine") == "curl"

    def test_translate_package_vim(self) -> None:
        """Test vim package translation."""
        assert translate_package("vim", "trixie") == "vim"
        assert translate_package("vim", "fedora") == "vim-enhanced"
        assert translate_package("vim", "alpine") == "vim"
        assert translate_package("vim", "arch") == "vim"

    def test_translate_package_iputils(self) -> None:
        """Test iputils-ping translation."""
        assert translate_package("iputils-ping", "trixie") == "iputils-ping"
        assert translate_package("iputils-ping", "fedora") == "iputils"
        assert translate_package("iputils-ping", "alpine") == "iputils"
        assert translate_package("iputils-ping", "arch") == "iputils"

    def test_translate_package_python(self) -> None:
        """Test Python package translation."""
        assert translate_package("python3-pip", "trixie") == "python3-pip"
        assert translate_package("python3-pip", "alpine") == "py3-pip"
        assert translate_package("python3-pip", "arch") == "python-pip"

    def test_translate_package_unknown(self) -> None:
        """Test unknown package returns as-is."""
        # Unknown packages should be returned unchanged
        assert translate_package("unknown-package", "trixie") == "unknown-package"

    def test_translate_packages_list(self) -> None:
        """Test translating a list of packages."""
        packages = ["vim", "curl", "iputils-ping"]
        translated = translate_packages(packages, "fedora")

        assert "vim-enhanced" in translated
        assert "curl" in translated
        assert "iputils" in translated
        assert len(translated) == 3

    def test_translate_packages_filters_none(self) -> None:
        """Test that unavailable packages are filtered out."""
        # Test with a package that might not exist on some distros
        packages = ["curl", "wget"]
        translated = translate_packages(packages, "trixie")
        assert len(translated) == 2


class TestServiceTranslation:
    """Tests for service name translation."""

    def test_translate_service_ssh(self) -> None:
        """Test SSH service translation."""
        assert translate_service("ssh", "trixie") == "ssh"
        assert translate_service("ssh", "fedora") == "sshd"
        assert translate_service("ssh", "alpine") == "sshd"
        assert translate_service("ssh", "arch") == "sshd"

    def test_translate_service_unknown(self) -> None:
        """Test unknown service returns as-is."""
        assert translate_service("custom-service", "trixie") == "custom-service"


class TestPackageManagerCommands:
    """Tests for package manager commands."""

    def test_apt_commands(self) -> None:
        """Test APT package manager commands."""
        commands = get_package_manager_commands("trixie")
        assert "apt-get update" in commands["update"]
        assert "apt-get install -y" in commands["install"]

    def test_dnf_commands(self) -> None:
        """Test DNF package manager commands."""
        commands = get_package_manager_commands("fedora")
        assert "dnf" in commands["update"]
        assert "dnf install -y" in commands["install"]

    def test_apk_commands(self) -> None:
        """Test APK package manager commands."""
        commands = get_package_manager_commands("alpine")
        assert "apk update" in commands["update"]
        assert "apk add" in commands["install"]

    def test_pacman_commands(self) -> None:
        """Test Pacman package manager commands."""
        commands = get_package_manager_commands("arch")
        assert "pacman -Sy" in commands["update"]
        assert "pacman -S --noconfirm" in commands["install"]


class TestServiceCommands:
    """Tests for service management commands."""

    def test_systemd_commands(self) -> None:
        """Test systemd service commands."""
        assert get_service_commands("trixie", "ssh", "enable") == "systemctl enable ssh"
        assert get_service_commands("trixie", "ssh", "start") == "systemctl start ssh"

    def test_systemd_with_translation(self) -> None:
        """Test systemd commands with service name translation."""
        assert get_service_commands("fedora", "ssh", "enable") == "systemctl enable sshd"
        assert get_service_commands("fedora", "ssh", "start") == "systemctl start sshd"

    def test_openrc_commands(self) -> None:
        """Test OpenRC service commands (Alpine)."""
        assert get_service_commands("alpine", "ssh", "enable") == "rc-update add sshd"
        assert get_service_commands("alpine", "ssh", "start") == "rc-service sshd start"
        assert get_service_commands("alpine", "ssh", "stop") == "rc-service sshd stop"


class TestImageAndBootstrap:
    """Tests for Docker images and bootstrap commands."""

    def test_get_docker_image(self) -> None:
        """Test getting Docker images."""
        assert get_docker_image("trixie") == "debian:trixie"
        assert get_docker_image("noble") == "ubuntu:noble"
        assert get_docker_image("fedora") == "fedora:41"
        assert get_docker_image("alpine") == "alpine:3.20"
        assert "archlinux" in get_docker_image("arch")

    def test_get_docker_image_unknown(self) -> None:
        """Test getting Docker image for unknown distro."""
        with pytest.raises(KeyError):
            get_docker_image("unknown")

    def test_get_cloud_image_url(self) -> None:
        """Test getting cloud image URLs."""
        url = get_cloud_image_url("trixie")
        assert "cloud.debian.org" in url

        url = get_cloud_image_url("noble")
        assert "cloud-images.ubuntu.com" in url

        url = get_cloud_image_url("fedora")
        assert "fedoraproject.org" in url

    def test_get_cloud_image_url_no_image(self) -> None:
        """Test getting cloud image for distro without one."""
        # Arch doesn't have official cloud images
        with pytest.raises(KeyError):
            get_cloud_image_url("arch")

    def test_get_debootstrap_cmd(self) -> None:
        """Test getting debootstrap commands."""
        cmd = get_debootstrap_cmd("trixie")
        assert "debootstrap" in cmd
        assert "trixie" in cmd

    def test_get_debootstrap_cmd_non_debian(self) -> None:
        """Test debootstrap not available for non-Debian distros."""
        with pytest.raises(KeyError):
            get_debootstrap_cmd("fedora")

        with pytest.raises(KeyError):
            get_debootstrap_cmd("alpine")

    def test_get_mirror(self) -> None:
        """Test getting package mirrors."""
        assert "debian.org" in get_mirror("trixie")
        assert "ubuntu.com" in get_mirror("noble")
        assert get_mirror("fedora") is None  # Fedora doesn't use mirrors this way
